"""Module to perform Lidar extraction and visualization of projections on image plane."""

import cv2
import numpy as np

from zod.constants import Camera, Lidar
from zod.data_classes.calibration import Calibration
from zod.data_classes.geometry import Pose
from zod.data_classes.sensor import LidarData
from zod.utils.geometry import get_points_in_camera_fov, project_3d_to_2d_kannala, transform_points
from zod.visualization.colorlabeler import ColorLabeler, create_matplotlib_colormap


def get_3d_transform_camera_lidar(
    calib: Calibration,
    lidar: Lidar = Lidar.VELODYNE,
    camera: Camera = Camera.FRONT,
) -> Pose:
    """Get 3D transformation between lidar and camera."""
    t_refframe_to_frame = calib.lidars[lidar].extrinsics
    t_refframe_from_frame = calib.cameras[camera].extrinsics

    t_from_frame_refframe = t_refframe_from_frame.inverse
    t_from_frame_to_frame = Pose(t_from_frame_refframe.transform @ t_refframe_to_frame.transform)

    return t_from_frame_to_frame


def project_lidar_to_image(
    lidar_data: LidarData,
    calib: Calibration,
    lidar: Lidar = Lidar.VELODYNE,
    camera: Camera = Camera.FRONT,
) -> np.ndarray:
    """Project lidar pointcloud to camera."""
    # TODO: use lidar param here
    t_lidar_to_camera = get_3d_transform_camera_lidar(calib)

    camera_data = transform_points(lidar_data.points, t_lidar_to_camera.transform)
    positive_depth = camera_data[:, 2] > 0
    camera_data = camera_data[positive_depth]
    if not camera_data.any():
        return camera_data
    camera_data = get_points_in_camera_fov(calib.cameras[camera].field_of_view, camera_data)

    xy_array = project_3d_to_2d_kannala(
        camera_data,
        calib.cameras[camera].intrinsics[..., :3],
        calib.cameras[camera].distortion,
    )
    xyd_array = np.concatenate([xy_array, camera_data[:, 2:3]], axis=1)
    return xyd_array


def draw_projections_as_points(
    image: np.ndarray, points: np.ndarray, clip_to: float = None
) -> np.ndarray:
    """Draw projected points from pointcloud to image plane as colored points.

    Args:
        image : image to draw projected points
        points: projected points from lidar
        clip_to : distance for clipping

    Returns:
        image : image with projected lidar points as colored points

    """
    color_labeler = ColorLabeler(map_type="coolwarm", map_creator=create_matplotlib_colormap)
    if points.shape[0]:
        normed_depths = points[:, 2] - points[:, 2].min()
        if clip_to:
            normed_depths = np.clip(normed_depths / clip_to, 0, 1)
        else:
            normed_depths = normed_depths / normed_depths.max()
        colors = [color_labeler.label_to_color_norm(d) for d in normed_depths]
        coords = points[:, :2].astype("int32")
        image[coords[:, 1], coords[:, 0]] = colors
    return image


def draw_projection_as_jet_circles(
    image: np.ndarray, points: np.ndarray, radius: int, clip_to: float = None
) -> np.ndarray:
    """Draw projected points from pointcloud to image plane as jet colored circles.

    Args:
        image : image to draw projected points
        points: projected points from lidar
        radius : radius of circle to be drawed
        clip_to : distance for clipping

    Returns:
        image : image with projected lidar points as colored circles with specified radius

    """
    color_labeler = ColorLabeler(map_type="coolwarm", map_creator=create_matplotlib_colormap)
    if points.shape[0]:
        normed_depths = points[:, 2] - points[:, 2].min()
        if clip_to:
            normed_depths = np.clip(normed_depths / clip_to, 0, 1)
        else:
            normed_depths = normed_depths / normed_depths.max()
        colors = [color_labeler.label_to_color_norm(d) for d in normed_depths]
        for point, color in zip(points[:, :2].astype("int32"), colors):
            center = tuple(point)
            cv2.circle(image, center, radius, color, -1)
    return image


def visualize_lidar_on_image(lidar_data: LidarData, calib: Calibration, image: np.ndarray):
    """Visualize GPS track on image."""
    xyd = project_lidar_to_image(lidar_data, calib)
    image = draw_projection_as_jet_circles(image, xyd, radius=2)
    return image



import json

from dataclass_wizard import JSONSerializable as _JSONSerializable


class JSONSerializable(_JSONSerializable):
    @classmethod
    def from_json_path(cls, json_path):
        with open(json_path) as fp:
            data = json.load(fp)
        if isinstance(data, dict):
            return cls.from_dict(data)
        elif isinstance(data, str):
            return cls.from_json(data)



from dataclasses import dataclass
from typing import Any, Dict, Optional, Union

import numpy as np
from pyquaternion import Quaternion

from zod.constants import EVALUATION_CLASSES, Camera, Lidar
from zod.data_classes.box import Box2D, Box3D


@dataclass
class AnnotatedObject:
    """Class to store dynamic object information."""

    # These are always available
    box2d: Box2D
    unclear: bool
    name: str
    uuid: str

    # These are not set if the object is unclear
    box3d: Optional[Box3D]  # This can be None even if the object is not unclear
    object_type: Optional[str]
    occlusion_level: Optional[str]
    artificial: Optional[str]
    with_rider: Optional[bool]
    emergency: Optional[bool]
    traffic_content_visible: Optional[bool]

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "AnnotatedObject":
        """Create an AnnotatedObject from a dictionary."""
        properties: Dict[str, Any] = data["properties"]

        box2d = Box2D.from_points(
            points=data["geometry"]["coordinates"],
            frame=Camera.FRONT,
        )
        if properties["class"] == "Inconclusive":
            unclear = True
        else:
            unclear = properties["unclear"]

        if "location_3d" not in properties:
            box3d = None
        else:
            box3d = Box3D(
                center=np.array(properties["location_3d"]["coordinates"]),
                size=np.array(
                    [
                        properties["size_3d_length"],
                        properties["size_3d_width"],
                        properties["size_3d_height"],
                    ]
                ),
                orientation=Quaternion(
                    properties["orientation_3d_qw"],
                    properties["orientation_3d_qx"],
                    properties["orientation_3d_qy"],
                    properties["orientation_3d_qz"],
                ),
                frame=Lidar.VELODYNE,
            )

        return cls(
            uuid=properties["annotation_uuid"],
            box2d=box2d,
            box3d=box3d,
            unclear=unclear,
            name=properties["class"],
            object_type=properties.get("type", None),
            occlusion_level=properties.get("occlusion_ratio", None),
            artificial=properties.get("artificial", None),
            with_rider=properties.get("with_rider", None) == "True",
            emergency=properties.get("emergency", None),
            traffic_content_visible=properties.get("traffic_content_visible", None) == "True",
        )

    def should_ignore_object(self, require_3d: bool = True, require_eval: bool = True) -> bool:
        """Check if the object should be ignored.

        Returns:
            True if the object should be ignored.
        """
        # Remove objects that are not to be evaluated
        if (self.name not in EVALUATION_CLASSES) and require_eval:
            return True
        # Remove unclear objects
        if self.unclear:
            return True
        # Remove objects that dont have 3d box
        if self.box3d is None and require_3d:
            return True
        # If the object is artificial, reflection or an image, ignore it
        if self.artificial not in (None, "None"):
            return True

        # Class specific removals
        if self.name == "Vehicle":
            if self.object_type not in (
                "Car",
                "Bus",
                "Truck",
                "Van",
                "Trailer",
                "HeavyEquip",
            ):
                return True
        elif self.name == "VulnerableVehicle":
            if self.object_type not in ("Bicycle", "Motorcycle"):
                return True
            return bool(self.with_rider)
        elif self.name == "Pedestrian":
            pass
        elif self.name == "TrafficSign":
            return not self.traffic_content_visible

        elif self.name == "TrafficSignal":
            return not self.traffic_content_visible

        return False


@dataclass
class PredictedObject:
    """Class to store dynamic object prediction information."""

    # TODO: maybe move to zod.eval?

    name: str
    confidence: float
    box3d: Box3D

    def __eq__(self, __o: Union[AnnotatedObject, "PredictedObject"]) -> bool:
        return self.box3d == __o.box3d



"""ZOD Object detection containers."""
from dataclasses import dataclass
from typing import List, Optional, Tuple

import numpy as np
from pyquaternion import Quaternion

from zod.constants import EGO, Camera, CoordinateFrame
from zod.data_classes import Calibration, Pose
from zod.utils.geometry import project_3d_to_2d_kannala, unproject_2d_to_3d_kannala


@dataclass
class Box3D:
    """Class to store a 3D bounding box.

    The box is defined by its center, size, orientation in the given coordinate frame.
    The size is defined as (length, width, height).
    """

    center: np.ndarray  # x, y, z
    size: np.ndarray  # L, W, H
    orientation: Quaternion
    frame: CoordinateFrame

    def _transform(self, transform: Optional[Pose], new_frame: CoordinateFrame):
        if transform is not None:
            self.center = transform.rotation_matrix @ self.center
            self.center += transform.translation
            self.orientation = transform.rotation * self.orientation
            self.frame = new_frame

    def _transform_inv(self, transform: Optional[Pose], new_frame: CoordinateFrame):
        if transform is not None:
            self.center -= transform.translation
            self.center = transform.rotation_matrix.T @ self.center
            self.orientation = transform.rotation.inverse * self.orientation
            self.frame = new_frame

    def convert_to(self, frame: CoordinateFrame, calib: Calibration):
        if frame == self.frame:
            return
        self._transform(calib.get_extrinsics(self.frame), new_frame=EGO)
        self._transform_inv(calib.get_extrinsics(frame), new_frame=frame)

    @property
    def corners(self) -> np.ndarray:
        """Get the corners of the bounding box in the current frame.

        Order of points are:
         - rear left bottom
         - rear right bottom
         - front right bottom
         - front left bottom
         - rear left top
         - rear right top
         - front right top
         - front left top
        """
        # Get the 3d corners of the box
        corners = np.array(
            [
                [-0.5, -0.5, -0.5],
                [0.5, -0.5, -0.5],
                [0.5, 0.5, -0.5],
                [-0.5, 0.5, -0.5],
                [-0.5, -0.5, 0.5],
                [0.5, -0.5, 0.5],
                [0.5, 0.5, 0.5],
                [-0.5, 0.5, 0.5],
            ]
        )
        corners *= self.size
        corners = self.orientation.rotation_matrix @ corners.T
        corners += self.center.reshape((-1, 1))
        return corners.T

    @property
    def corners_bev(self) -> np.ndarray:
        """Get the corners of the 3D bounding box where the z-dimension is (birds-eye-view)."""
        return self.corners[:4, :2]

    def project_into_camera(self, calib: Calibration) -> np.ndarray:
        """Project 3D bounding box into 2D image plane.

        Return an array (9, 2) the box center and the box corners in the imageplane
        of the given camera. The order of the points is:
            - center point
            - rear left bottom
            - rear right bottom
            - front right bottom
            - front left bottom
            - rear left top
            - rear right top
            - front right top
            - front left top
        """
        if not isinstance(self.frame, Camera):
            raise ValueError(f"Cannot project box into frame {self.frame}")

        # Concatenate the corners and center of the box
        points = np.concatenate([self.center.reshape((1, -1)), self.corners], axis=0)

        # Project center into camera
        pos2d = project_3d_to_2d_kannala(
            points,
            calib.cameras[self.frame].intrinsics,
            calib.cameras[self.frame].distortion,
        )
        return pos2d

    def __eq__(self, __o: "Box3D") -> bool:
        b1 = np.allclose(self.center, __o.center)
        b2 = np.allclose(self.size, __o.size)
        b3 = self.orientation == __o.orientation
        b4 = self.frame == __o.frame
        return b1 and b2 and b3 and b4


@dataclass
class Box2D:
    """2D Bounding Box container."""

    xyxy: np.ndarray  # xmin, ymin, xmax, ymax
    frame: Camera

    @classmethod
    def from_points(cls, points: List[List[float]], frame=Camera.FRONT) -> "Box2D":
        """Compute outer points from a polygon.

        Args:
            points (list): list of points shaping bounding box and the properties.

        """
        xmin = ymin = 999999
        xmax = ymax = -1
        for (point_x, point_y) in points:
            xmin = min(xmin, point_x)
            xmax = max(xmax, point_x)
            ymin = min(ymin, point_y)
            ymax = max(ymax, point_y)
        return cls(xyxy=np.array([xmin, ymin, xmax, ymax], dtype=np.float32), frame=frame)

    @property
    def xywh(self) -> np.ndarray:
        """Get the box in xywh format."""
        return self.xyxy - np.array([0, 0, self.xyxy[0], self.xyxy[1]], dtype=self.xyxy.dtype)

    @property
    def area(self) -> float:
        """Bounding Box 2D area.

        Returns:
            Area of the Bounding Box.
        """
        return float(np.prod(self.xyxy[2:] - self.xyxy[:2]))

    @property
    def center(self) -> np.ndarray:
        """Bounding Box 2D center.

        Returns:
            Center of the Bounding Box.
        """
        return (self.xyxy[:2] + self.xyxy[2:]) / 2

    @property
    def dimension(self) -> np.ndarray:
        """Bounding Box 2D dimension.

        Returns:
            Dimension of the Bounding Box.
        """
        return self.xyxy[2:] - self.xyxy[:2]

    @property
    def corners(
        self,
    ) -> np.ndarray:
        """Bounding Box 2D corners.

        Returns:
            array (4, 2) of the corners of the Bounding Box.
            - top left
            - top right
            - bottom right
            - bottom left
        """
        return np.stack(
            [
                self.xyxy[[0, 1]],
                self.xyxy[[2, 1]],
                self.xyxy[[2, 3]],
                self.xyxy[[0, 3]],
            ]
        )

    @property
    def xmin(self) -> int:
        """Get the (rounded) left pixel coordinate."""
        return int(self.xyxy[0].round())

    @property
    def ymin(self) -> int:
        """Get the (rounded) top pixel coordinate."""
        return int(self.xyxy[1].round())

    @property
    def xmax(self) -> int:
        """Get the (rounded) right pixel coordinate."""
        return int(self.xyxy[2].round())

    @property
    def ymax(self) -> int:
        """Get the (rounded) bottom pixel coordinate."""
        return int(self.xyxy[3].round())

    def crop_from_image(
        self,
        image: np.ndarray,
        padding: Optional[Tuple[int, int]],
        padding_factor: Optional[float],
    ) -> Tuple[np.ndarray, Tuple[int, int, int, int]]:
        """Crop out the Bounding Box from an image.

        Args:
            image: Image to crop out the Bounding Box from.
            padding: Tuple of padding in x and y direction.
            padding_factor: Factor to multiply the padding with.

        Returns:
            Cropped image.
            tuple: padding in all four directions | (left,  top,  right, bottom)
        """

        assert not (
            padding is not None and padding_factor is not None
        ), "Cannot specify both padding and padding_factor"

        if padding is not None:
            padding_x, padding_y = padding
        elif padding_factor is not None:
            padding_x = int(self.dimension[0] * padding_factor)
            padding_y = int(self.dimension[1] * padding_factor)
        else:
            padding_x = 0
            padding_y = 0

        xmin = max(0, self.xmin - padding_x)
        ymin = max(0, self.ymin - padding_y)
        xmax = min(image.shape[1], self.xmax + padding_x)
        ymax = min(image.shape[0], self.ymax + padding_y)

        return image[ymin:ymax, xmin:xmax], (
            self.xmin - xmin,
            self.ymin - ymin,
            xmax - self.xmax,
            ymax - self.ymax,
        )

    def points_in_box(self, points: np.ndarray) -> np.ndarray:
        """Check if a point is inside the Bounding Box.

        Args:
            point: points to check. (n, 2)

        Returns:
            True if point is inside the Bounding Box.
        """

        return points >= self.xyxy[:2] & points <= self.xyxy[2:]

    def get_3d_frustum(
        self,
        calibration: Calibration,
        frame: Optional[CoordinateFrame] = None,
        min_depth: float = 0.0,
        max_depth: float = 500.0,
    ) -> np.ndarray:
        """Get the frustum of the bounding box in homogenous coordinates.

        The frustum is defined by the by having the 4 corners of the bounding box
        projected to the max_depth along with camera center (0, 0, 0).
        Args:
            calibration: Calibration data.
            frame: Frame to project the frustum into.
            max_depth: maximum depth of the frustum.

        Returns:
            frustum: (8, 4) array of the frustum. Note that is min_depth == 0.0,
            the first 4 points will correspond to camera center (0, 0, 0)
        """
        # Get the 2d bounding box corners
        corners = self.corners
        camera_calib = calibration.cameras[self.frame]
        if min_depth > 0:
            # Copy the corners
            corners = np.concatenate([corners, corners], axis=0)

            # Create the depth vector where the first four correspond to min_depth
            # and the last four correspond to max_depth
            depth = np.array([min_depth] * 4 + [max_depth] * 4)

        else:
            depth = np.array([max_depth] * 4)

        # Project the 2d corners to the max_depth using the calibration
        frustum = unproject_2d_to_3d_kannala(
            corners, camera_calib.intrinsics, camera_calib.undistortion, depth
        )

        if min_depth == 0.0:
            frustum = np.concatenate((np.zeros((4, 3)), frustum), axis=0)

        # Transform the frustum to the selected frame if needed
        if frame is not None:
            frustum = calibration.transform_points(
                points=frustum,
                from_frame=self.frame,
                to_frame=frame,
            )

        return frustum



"""This script is to be used to download the zenseact open dataset."""
import contextlib
import os
import os.path as osp
import subprocess
import tarfile
import threading
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass

from tqdm import tqdm

try:
    import click.exceptions
    import dropbox
    import dropbox.exceptions
    import dropbox.files
    import dropbox.sharing
    import typer
except ImportError:
    print('zod is installed without the CLI dependencies: pip install "zod[cli]"')
    exit(1)


APP_KEY = "kcvokw7c7votepo"
REFRESH_TOKEN = "z2h_5rjphJEAAAAAAAAAAUWtQlltCYlMbZ2WqMgtymELhiSuKIksxAYeArubKnZV"
CHUNK_SIZE = 1024 * 1024  # 1 MB
TIMEOUT = 60 * 60  # 1 hour

app = typer.Typer(help="Zenseact Open Dataset Donwloader", no_args_is_help=True)


@dataclass
class ExtractInfo:
    """Information about a file to extract."""

    url: str
    file_path: str
    output_dir: str
    rm: bool
    dry_run: bool
    size: int
    extract: bool


@dataclass
class FilterSettings:
    """Filter settings."""

    mini: bool
    annotations: bool
    images: bool
    blur: bool
    dnat: bool
    lidar: bool
    oxts: bool
    calibrations: bool
    num_scans_before: int
    num_scans_after: int


@dataclass
class DownloadSettings:
    """Download settings."""

    url: str
    output_dir: str
    rm: bool
    dry_run: bool
    extract: bool
    parallel: bool


class ResumableDropbox(dropbox.Dropbox):
    """A patched dropbox client that allows to resume downloads."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._headers = {} if self._headers is None else self._headers
        self._lock = threading.Lock()

    def sharing_get_shared_link_file(self, url, path=None, link_password=None, start=0):
        """
        Download the shared link's file from a user's Dropbox.
        """
        # Need to lock here because the headers are shared between all requests
        with self._lock:
            self._headers["Range"] = f"bytes={start}-"
            res = super().sharing_get_shared_link_file(url, path=path, link_password=link_password)
            del self._headers["Range"]
        return res


def _print_final_msg(pbar: tqdm, basename: str):
    """Print a final message for each downloaded file, with stats and nice padding."""
    msg = "Downloaded " + basename + " " * (45 - len(basename))

    # Print various download stats
    total_time = pbar.format_dict["elapsed"]
    size_in_mb = pbar.n / 1024 / 1024
    speed = size_in_mb / total_time
    for name, val, unit in [
        ("time", total_time, "s"),
        ("size", size_in_mb, "MB"),
        ("speed", speed, "MB/s"),
    ]:
        val = f"{val:.2f}{unit}"
        msg += f"{name}: {val}" + " " * (14 - len(val))

    tqdm.write(msg)


def _download(download_path: str, dbx: ResumableDropbox, info: ExtractInfo):
    current_size = 0
    if osp.exists(download_path):
        current_size = osp.getsize(download_path)
    basename = osp.basename(info.file_path)
    pbar = tqdm(
        total=info.size,
        unit="iB",
        unit_scale=True,
        desc=f"Downloading {basename}...",
        leave=False,
        initial=current_size,
    )
    # Retry download if partial file exists (e.g. due to network error)
    while pbar.n < info.size:
        if pbar.n > 0:
            # this means we are retrying or resuming a previously interrupted download
            tqdm.write(f"Resuming download of {download_path} from {current_size} bytes.")
        _, response = dbx.sharing_get_shared_link_file(
            url=info.url, path=info.file_path, start=pbar.n
        )
        with open(download_path, "ab") as f:
            with contextlib.closing(response):
                for chunk in response.iter_content(chunk_size=CHUNK_SIZE):
                    if chunk:  # filter out keep-alive new chunks
                        size = f.write(chunk)
                        pbar.update(size)
    # Final checks and printouts
    assert osp.getsize(download_path) == info.size
    _print_final_msg(pbar, basename)


def _extract(tar_path: str, output_dir: str):
    """Extract a tar file to a directory."""
    pbar = tqdm(desc=f"Extracting {osp.basename(tar_path)}...", leave=False)
    # Check if system has tar installed (pipe stderr to None to avoid printing)
    if os.system("tar --version > /dev/null 2>&1") == 0:
        with subprocess.Popen(
            ["tar", "-xvf", tar_path, "-C", output_dir],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        ) as tar:
            for line in tar.stdout:
                pbar.update(1)
        # check if tar exited with error
        if tar.returncode != 0:
            tqdm.write(f"Error extracting {tar_path} with tar.")
    else:
        # Fallback to python tarfile
        with tarfile.open(name=tar_path) as tar:
            # pbar.total = len(tar.getmembers())
            for member in tar.getmembers():
                tar.extract(member=member, path=output_dir)
                pbar.update(1)
    tqdm.write(f"Extracted {pbar.n} files from {osp.basename(tar_path)}.")


def _download_and_extract(dbx: ResumableDropbox, info: ExtractInfo):
    """Download a file from a Dropbox share link to a local path."""
    download_path = osp.join(info.output_dir, "downloads", osp.basename(info.file_path))
    if osp.exists(download_path) and osp.getsize(download_path) == info.size:
        tqdm.write(f"File {download_path} already exists. Skipping download.")
    elif info.dry_run:
        typer.echo(f"Would download and extract {info.file_path} to {download_path}")
        return
    else:
        _download(download_path, dbx, info)

    if info.extract:
        if info.dry_run:
            typer.echo(f"Would extract {download_path} to {info.output_dir}")
            return
        else:
            _extract(download_path, info.output_dir)

    if info.rm and not info.dry_run:
        os.remove(download_path)


def _filter_entry(entry: dropbox.files.Metadata, settings: FilterSettings) -> bool:
    """Filter the entry based on the flags."""
    if settings.mini and "mini" not in entry.name:
        # If mini is set, we only want to download the mini dataset
        return False
    # If we are downloading the full dataset, we don't want the mini dataset
    if not settings.mini and "mini" in entry.name:
        return False

    if not settings.annotations and "annotations" in entry.name:
        return False
    if "images" in entry.name:
        if not settings.images:
            return False
        if not settings.blur and "blur" in entry.name:
            return False
        if not settings.dnat and "dnat" in entry.name:
            return False
    if not settings.oxts and "oxts" in entry.name:
        return False
    if not settings.calibrations and "calibrations" in entry.name:
        return False
    if "lidar" in entry.name:
        if not settings.lidar:
            return False
        distance = entry.name.split("_")[2][:2]
        if "after" in entry.name and (int(distance) > settings.num_scans_after):
            return False
        if "before" in entry.name and (int(distance) > settings.num_scans_before):
            return False
    return True


def _download_dataset(dl_settings: DownloadSettings, filter_settings: FilterSettings, dirname: str):
    dbx = ResumableDropbox(app_key=APP_KEY, oauth2_refresh_token=REFRESH_TOKEN, timeout=TIMEOUT)
    url, entries = _list_folder(dl_settings.url, dbx, dirname)
    files_to_download = [
        ExtractInfo(
            file_path=f"/{dirname}/" + entry.name,
            size=entry.size,
            url=url,
            output_dir=dl_settings.output_dir,
            rm=dl_settings.rm,
            dry_run=dl_settings.dry_run,
            extract=dl_settings.extract,
        )
        for entry in entries
        if _filter_entry(entry, filter_settings)
    ]
    if not files_to_download:
        typer.echo("No files to download. Perhaps you specified too many filters?")
        return

    if not dl_settings.dry_run:
        os.makedirs(dl_settings.output_dir, exist_ok=True)
        os.makedirs(osp.join(dl_settings.output_dir, "downloads"), exist_ok=True)
    if dl_settings.parallel:
        with ThreadPoolExecutor() as pool:
            pool.map(lambda info: _download_and_extract(dbx, info), files_to_download)
    else:
        for info in files_to_download:
            _download_and_extract(dbx, info)


def _list_folder(url, dbx, path):
    shared_link = dropbox.files.SharedLink(url=url)
    try:
        res = dbx.files_list_folder(path=f"/{path}", shared_link=shared_link)
    except dropbox.exceptions.ApiError as err:
        raise click.exceptions.ClickException(
            f"Dropbox raised the following error:\n\t{err}\nThis could be due to a bad url."
        )

    entries = res.entries
    while res.has_more:
        res = dbx.files_list_folder_continue(res.cursor)
        entries.extend(res.entries)
    return url, entries


@app.callback(no_args_is_help=True)
def common(
    ctx: typer.Context,
    url: str = typer.Option(..., help="The dropbox shared folder url"),
    output_dir: str = typer.Option(..., help="The output directory"),
    rm: bool = typer.Option(False, help="Whether to remove the downloaded archives"),
    dry_run: bool = typer.Option(
        False, help="Whether to only print the files that would be downloaded"
    ),
    extract: bool = typer.Option(True, help="Whether to unpack the archives"),
    parallel: bool = typer.Option(True, help="Whether to download files in parallel"),
):
    ctx.obj = DownloadSettings(
        url=url,
        output_dir=output_dir,
        rm=rm,
        dry_run=dry_run,
        extract=extract,
        parallel=parallel,
    )


@app.command()
def frames(
    ctx: typer.Context,
    mini: bool = typer.Option(False, help="Whether to download the mini dataset"),
    annotations: bool = typer.Option(True, help="Whether to download the annotations"),
    images: bool = typer.Option(True, help="Whether to download the images"),
    blur: bool = typer.Option(True, help="Whether to download the blur images"),
    dnat: bool = typer.Option(False, help="Whether to download the dnat images"),
    lidar: bool = typer.Option(True, help="Whether to download the lidar data"),
    num_scans_before: int = typer.Option(
        0, help="Number of earlier lidar scans to download (-1 == all)"
    ),
    num_scans_after: int = typer.Option(
        0, help="Number of later lidar scans to download (-1 == all)"
    ),
    oxts: bool = typer.Option(True, help="Whether to download the oxts data"),
    calibrations: bool = typer.Option(True, help="Whether to download the calibration data"),
):
    """Download the zenseact open dataset."""

    if images:
        assert blur or dnat, "Must download at least one type of image"

    dl_settings: DownloadSettings = ctx.obj
    filter_settings = FilterSettings(
        mini=mini,
        annotations=annotations,
        images=images,
        blur=blur,
        dnat=dnat,
        lidar=lidar,
        oxts=oxts,
        calibrations=calibrations,
        num_scans_before=num_scans_before,
        num_scans_after=num_scans_after,
    )
    _download_dataset(dl_settings, filter_settings, "single_frames")


@app.command()
def sequences(
    ctx: typer.Context,
    mini: bool = typer.Option(False, help="Whether to download the mini dataset"),
    annotations: bool = typer.Option(True, help="Whether to download the annotations"),
    images: bool = typer.Option(True, help="Whether to download the images"),
    lidar: bool = typer.Option(True, help="Whether to download the lidar data"),
    oxts: bool = typer.Option(True, help="Whether to download the oxts data"),
    calibrations: bool = typer.Option(True, help="Whether to download the calibration data"),
):
    """Download the zenseact open dataset."""
    dl_settings: DownloadSettings = ctx.obj
    filter_settings = FilterSettings(
        mini=mini,
        annotations=annotations,
        images=images,
        blur=True,
        dnat=False,
        lidar=lidar,
        oxts=oxts,
        calibrations=calibrations,
        num_scans_before=-1,
        num_scans_after=-1,
    )
    _download_dataset(dl_settings, filter_settings, "sequences")


if __name__ == "__main__":
    app()



"""ZOD dataclasses."""
from dataclasses import dataclass
from datetime import datetime
from typing import Any, List, Union

import numpy as np
from PIL import Image

from zod.anno.parser import ANNOTATION_PARSERS
from zod.constants import AnnotationProject
from zod.utils.utils import parse_datetime_from_filename

from ._serializable import JSONSerializable
from .geometry import Pose


@dataclass
class LidarData:
    """A class describing the lidar data."""

    points: np.ndarray  # (N, 3) float32
    timestamps: np.ndarray  # (N,) float64 epoch timestamp pointwise
    intensity: np.ndarray  # (N,) uint8
    diode_idx: np.ndarray  # (N,) uint8
    core_timestamp: float  # core epoch timestamp

    def copy(self) -> "LidarData":
        """Return a copy of the lidar data."""
        return LidarData(
            points=self.points.copy(),
            timestamps=self.timestamps.copy(),
            intensity=self.intensity.copy(),
            diode_idx=self.diode_idx.copy(),
            core_timestamp=self.core_timestamp,
        )

    @classmethod
    def empty(cls) -> "LidarData":
        """Create an empty lidar data object."""
        return cls(
            points=np.empty((0, 3), dtype=np.float32),
            timestamps=np.empty(0, dtype=np.float64),
            intensity=np.empty(0, dtype=np.uint8),
            diode_idx=np.empty(0, dtype=np.uint8),
            core_timestamp=0.0,
        )

    @classmethod
    def from_npy(cls, path: str) -> "LidarData":
        """Load lidar data from a .npy file."""
        data = np.load(path)
        core_timestamp = parse_datetime_from_filename(path).timestamp()
        return cls(
            points=np.vstack((data["x"], data["y"], data["z"])).T,
            timestamps=core_timestamp + data["timestamp"] / 1e6,
            intensity=data["intensity"],
            diode_idx=data["diode_index"],
            core_timestamp=core_timestamp,
        )

    def to_npy(self, path: str) -> None:
        """Save lidar data to a .npy file in the same format as is used for loading."""
        data = np.empty(
            len(self.points),
            dtype=[
                ("x", np.float32),
                ("y", np.float32),
                ("z", np.float32),
                ("timestamp", np.int64),
                ("intensity", np.uint8),
                ("diode_index", np.uint8),
            ],
        )
        data["x"] = self.points[:, 0]
        data["y"] = self.points[:, 1]
        data["z"] = self.points[:, 2]
        data["timestamp"] = (self.timestamps - self.core_timestamp) * 1e6
        data["intensity"] = self.intensity
        data["diode_index"] = self.diode_idx
        np.save(path, data)

    def transform(self, pose: Union[np.ndarray, Pose]) -> None:
        """Transform the lidar data to a new pose.

        Args:
            pose: The new pose to transform the lidar data to.
        """
        if isinstance(pose, Pose):
            pose = pose.transform
        rotations = pose[..., :3, :3]
        translations = pose[..., :3, 3]
        self.points = (
            self.points[..., None, :] @ rotations.swapaxes(-2, -1) + translations[..., None, :]
        ).squeeze(-2)

    def extend(self, *other: "LidarData"):
        """Extend this LidarData with data from another LidarData object.

        Args:
            other: The other LidarData object.
        """
        self.points = np.vstack((self.points, *(o.points for o in other)))
        self.timestamps = np.hstack((self.timestamps, *(o.timestamps for o in other)))
        self.intensity = np.hstack((self.intensity, *(o.intensity for o in other)))
        self.diode_idx = np.hstack((self.diode_idx, *(o.diode_idx for o in other)))
        # Core timestamp is the weighted average
        self.core_timestamp = (
            self.core_timestamp * len(self.timestamps)
            + sum(o.core_timestamp * len(o.timestamps) for o in other)
        ) / (len(self.timestamps) + sum(len(o.timestamps) for o in other))

    def __eq__(self, other: "LidarData") -> bool:
        """Check if two LidarData objects are equal."""
        return (
            np.allclose(self.points, other.points)
            and np.allclose(self.timestamps, other.timestamps)
            and np.allclose(self.intensity, other.intensity)
            and np.allclose(self.diode_idx, other.diode_idx)
            and np.isclose(self.core_timestamp, other.core_timestamp)
        )


@dataclass
class SensorFrame(JSONSerializable):
    """Class to store sensor information."""

    filepath: str
    time: datetime

    def read(self) -> Any:
        """Read the data from the file."""
        raise NotImplementedError


@dataclass
class LidarFrame(SensorFrame):
    """Class to store information about a lidar frame."""

    is_compensated: bool  # Whether the cloud is pointwise compensated

    def read(self) -> LidarData:
        """Read the point cloud."""
        return LidarData.from_npy(self.filepath)


@dataclass
class CameraFrame(SensorFrame):
    """Class to store information about a camera frame."""

    height: int = 2168
    width: int = 3848

    def read(self) -> np.ndarray:
        """Read the image."""
        return np.array(Image.open(self.filepath))


@dataclass
class AnnotationFrame(SensorFrame):
    """Class to store information about an annotation frame."""

    project: AnnotationProject

    def read(self) -> List[Any]:
        """Read (and parse) the annotation json."""
        return ANNOTATION_PARSERS[self.project](self.filepath)



import cv2
import numpy as np

from zod.utils.polygon_transformations import polygon_to_mask


def overlay_polygon_on_image(polygon, image, fill_color=(100, 0, 0), alpha=1.0):
    img_shape = np.shape(image)
    out_img = image.copy()

    image_mask = polygon_to_mask(polygon, img_shape, fill_color)

    out_img = cv2.addWeighted(out_img, 1, image_mask, alpha, -1)

    return out_img


def overlay_mask_on_image(mask, image, fill_color=(100, 0, 0), alpha=1.0):
    img_shape = np.shape(image)
    out_img = image.copy()
    mask_img = np.zeros(img_shape, dtype=np.uint8)

    for i, c in enumerate(fill_color):
        mask_img[:, :, i] = mask.astype(np.uint8) * np.uint8(c)

    out_img = cv2.addWeighted(out_img, 1, mask_img, alpha, -1)

    return out_img



"""Annotation parsers."""
import json
from typing import Any, Dict, List

from zod.anno.tsr.class_map import get_class_idx
from zod.constants import AnnotationProject, Camera
from zod.data_classes.box import Box2D

from .object import AnnotatedObject
from .tsr.traffic_sign import TrafficSignAnnotation


def _read_annotation_file(annotation_file: str) -> List[Dict[str, Any]]:
    """Read an annotation file."""
    with open(annotation_file, "r") as file:
        annotation = json.load(file)
    return annotation


def parse_object_detection_annotation(annotation_path: str) -> List[AnnotatedObject]:
    """Parse the objects annotation from the annotation string."""
    annotation = _read_annotation_file(annotation_path)
    return [AnnotatedObject.from_dict(obj) for obj in annotation]


def parse_traffic_sign_annotation(annotation_path: str) -> List[TrafficSignAnnotation]:
    """Parse the traffic sign annotation from the annotation string."""
    annotation = _read_annotation_file(annotation_path)
    annotated_objects = []
    for annotated_object in annotation:
        # Ignore all unclear traffic signs
        if annotated_object["properties"]["unclear"]:
            continue
        bounding_box = Box2D.from_points(annotated_object["geometry"]["coordinates"], Camera.FRONT)
        annotated_objects.append(
            TrafficSignAnnotation(
                bounding_box=bounding_box,
                traffic_sign_class=annotated_object["properties"]["class"],
                traffic_sign_idx=get_class_idx(annotated_object["properties"]["class"]),
                occlusion_ratio=annotated_object["properties"]["occlusion_ratio"],
                annotation_uuid=annotated_object["properties"]["annotation_uuid"],
                electronic_sign=annotated_object["properties"]["is_electronic"],
                uuid=annotated_object["properties"]["annotation_uuid"],
            )
        )
    return annotated_objects


def parse_lane_markings_annotation(annotation_path: str, classes=["lm_dashed", "lm_solid"]):
    """Parse the lane markings annotation from the annotation string."""
    annotations = _read_annotation_file(annotation_path)
    polygons = []
    for annotation in annotations:
        if "class" in annotation["properties"]:
            annotated_class = annotation["properties"]["class"]
            if annotated_class in classes:
                polygons.append(annotation["geometry"]["coordinates"])
    return polygons


def parse_ego_road_annotation(annotation_path: str, classes=["EgoRoad_Road"]):
    """Parse the egoroad annotation from the annotation string."""
    annotations = _read_annotation_file(annotation_path)
    polygons = []
    for annotation in annotations:
        if "class" in annotation["properties"]:
            annotated_class = annotation["properties"]["class"]
            if annotated_class in classes:
                polygons.append(annotation["geometry"]["coordinates"])
    return polygons


ANNOTATION_PARSERS = {
    AnnotationProject.LANE_MARKINGS: parse_lane_markings_annotation,
    AnnotationProject.EGO_ROAD: parse_ego_road_annotation,
    AnnotationProject.OBJECT_DETECTION: parse_object_detection_annotation,
    AnnotationProject.TRAFFIC_SIGNS: parse_traffic_sign_annotation,
}


if __name__ == "__main__":
    pass



# nuScenes dev-kit.
# Code written by Holger Caesar, 2018.

from typing import Any

import numpy as np
from pyquaternion import Quaternion

from ..common.data_classes import EvalBox

DetectionBox = Any  # Workaround as direct imports lead to cyclic dependencies.


def center_distance(gt_box: EvalBox, pred_box: EvalBox) -> float:
    """
    L2 distance between the box centers (xy only).
    :param gt_box: GT annotation sample.
    :param pred_box: Predicted sample.
    :return: L2 distance.
    """
    return np.linalg.norm(np.array(pred_box.translation[:2]) - np.array(gt_box.translation[:2]))


def velocity_l2(gt_box: EvalBox, pred_box: EvalBox) -> float:
    """
    L2 distance between the velocity vectors (xy only).
    If the predicted velocities are nan, we return inf, which is subsequently clipped to 1.
    :param gt_box: GT annotation sample.
    :param pred_box: Predicted sample.
    :return: L2 distance.
    """
    return np.linalg.norm(np.array(pred_box.velocity) - np.array(gt_box.velocity))


def yaw_diff(gt_box: EvalBox, eval_box: EvalBox, period: float = 2 * np.pi) -> float:
    """
    Returns the yaw angle difference between the orientation of two boxes.
    :param gt_box: Ground truth box.
    :param eval_box: Predicted box.
    :param period: Periodicity in radians for assessing angle difference.
    :return: Yaw angle difference in radians in [0, pi].
    """
    yaw_gt = quaternion_yaw(Quaternion(gt_box.rotation))
    yaw_est = quaternion_yaw(Quaternion(eval_box.rotation))

    return abs(angle_diff(yaw_gt, yaw_est, period))


def angle_diff(x: float, y: float, period: float) -> float:
    """
    Get the smallest angle difference between 2 angles: the angle from y to x.
    :param x: To angle.
    :param y: From angle.
    :param period: Periodicity in radians for assessing angle difference.
    :return: <float>. Signed smallest between-angle difference in range (-pi, pi).
    """

    # calculate angle difference, modulo to [0, 2*pi]
    diff = (x - y + period / 2) % period - period / 2
    if diff > np.pi:
        diff = diff - (2 * np.pi)  # shift (pi, 2*pi] to (-pi, 0]

    return diff


def attr_acc(gt_box: DetectionBox, pred_box: DetectionBox) -> float:
    """
    Computes the classification accuracy for the attribute of this class (if any).
    If the GT class has no attributes or the annotation is missing attributes, we assign an accuracy of nan, which is
    ignored later on.
    :param gt_box: GT annotation sample.
    :param pred_box: Predicted sample.
    :return: Attribute classification accuracy (0 or 1) or nan if GT annotation does not have any attributes.
    """
    if gt_box.attribute_name == "":
        # If the class does not have attributes or this particular sample is missing attributes, return nan, which is
        # ignored later. Note that about 0.4% of the sample_annotations have no attributes, although they should.
        acc = np.nan
    else:
        # Check that label is correct.
        acc = float(gt_box.attribute_name == pred_box.attribute_name)
    return acc


def scale_iou(sample_annotation: EvalBox, sample_result: EvalBox) -> float:
    """
    This method compares predictions to the ground truth in terms of scale.
    It is equivalent to intersection over union (IOU) between the two boxes in 3D,
    if we assume that the boxes are aligned, i.e. translation and rotation are considered identical.
    :param sample_annotation: GT annotation sample.
    :param sample_result: Predicted sample.
    :return: Scale IOU.
    """
    # Validate inputs.
    sa_size = np.array(sample_annotation.size)
    sr_size = np.array(sample_result.size)
    assert all(sa_size > 0), "Error: sample_annotation sizes must be >0."
    assert all(sr_size > 0), "Error: sample_result sizes must be >0."

    # Compute IOU.
    min_wlh = np.minimum(sa_size, sr_size)
    volume_annotation = np.prod(sa_size)
    volume_result = np.prod(sr_size)
    intersection = np.prod(min_wlh)  # type: float
    union = volume_annotation + volume_result - intersection  # type: float
    iou = intersection / union

    return iou


def quaternion_yaw(q: Quaternion) -> float:
    """
    Calculate the yaw angle from a quaternion.
    Note that this only works for a quaternion that represents a box in lidar or global coordinate frame.
    It does not work for a box in the camera frame.
    :param q: Quaternion of interest.
    :return: Yaw angle in radians.
    """

    # Project into xy plane.
    v = np.dot(q.rotation_matrix, np.array([1, 0, 0]))

    # Measure yaw using arctan.
    yaw = np.arctan2(v[1], v[0])

    return yaw


# def boxes_to_sensor(boxes: List[EvalBox], pose_record: Dict, cs_record: Dict):
#     """
#     Map boxes from global coordinates to the vehicle's sensor coordinate system.
#     :param boxes: The boxes in global coordinates.
#     :param pose_record: The pose record of the vehicle at the current timestamp.
#     :param cs_record: The calibrated sensor record of the sensor.
#     :return: The transformed boxes.
#     """
#     boxes_out = []
#     for box in boxes:
#         # Create Box instance.
#         box = Box(box.translation, box.size, Quaternion(box.rotation))

#         # Move box to ego vehicle coord system.
#         box.translate(-np.array(pose_record["translation"]))
#         box.rotate(Quaternion(pose_record["rotation"]).inverse)

#         #  Move box to sensor coord system.
#         box.translate(-np.array(cs_record["translation"]))
#         box.rotate(Quaternion(cs_record["rotation"]).inverse)

#         boxes_out.append(box)

#     return boxes_out


def cummean(x: np.array) -> np.array:
    """
    Computes the cumulative mean up to each position in a NaN sensitive way
    - If all values are NaN return an array of ones.
    - If some values are NaN, accumulate arrays discording those entries.
    """
    if sum(np.isnan(x)) == len(x):
        # Is all numbers in array are NaN's.
        return np.ones(len(x))  # If all errors are NaN set to error to 1 for all operating points.
    else:
        # Accumulate in a nan-aware manner.
        sum_vals = np.nancumsum(x.astype(float))  # Cumulative sum ignoring nans.
        count_vals = np.cumsum(~np.isnan(x))  # Number of non-nans up to each position.
        return np.divide(sum_vals, count_vals, out=np.zeros_like(sum_vals), where=count_vals != 0)



# nuScenes dev-kit.
# Code written by Oscar Beijbom, 2019.

from typing import Callable

import numpy as np

from ..common.data_classes import EvalBoxes
from ..common.utils import attr_acc, center_distance, cummean, scale_iou, velocity_l2, yaw_diff
from ..detection.data_classes import DetectionMetricData


def accumulate(
    gt_boxes: EvalBoxes,
    pred_boxes: EvalBoxes,
    class_name: str,
    dist_fcn: Callable,
    dist_th: float,
    verbose: bool = False,
) -> DetectionMetricData:
    """
    Average Precision over predefined different recall thresholds for a single distance threshold.
    The recall/conf thresholds and other raw metrics will be used in secondary metrics.
    :param gt_boxes: Maps every sample_token to a list of its sample_annotations.
    :param pred_boxes: Maps every sample_token to a list of its sample_results.
    :param class_name: Class to compute AP on.
    :param dist_fcn: Distance function used to match detections and ground truths.
    :param dist_th: Distance threshold for a match.
    :param verbose: If true, print debug messages.
    :return: (average_prec, metrics). The average precision value and raw data for a number of metrics.
    """
    # ---------------------------------------------
    # Organize input and initialize accumulators.
    # ---------------------------------------------

    # Count the positives.
    npos = len([1 for gt_box in gt_boxes.all if gt_box.detection_name == class_name])
    if verbose:
        print(
            "Found {} GT of class {} out of {} total across {} samples.".format(
                npos, class_name, len(gt_boxes.all), len(gt_boxes.sample_tokens)
            )
        )

    # For missing classes in the GT, return a data structure corresponding to no predictions.
    if npos == 0:
        return DetectionMetricData.no_predictions()

    # Organize the predictions in a single list.
    pred_boxes_list = [box for box in pred_boxes.all if box.detection_name == class_name]
    pred_confs = [box.detection_score for box in pred_boxes_list]

    if verbose:
        print(
            "Found {} PRED of class {} out of {} total across {} samples.".format(
                len(pred_confs),
                class_name,
                len(pred_boxes.all),
                len(pred_boxes.sample_tokens),
            )
        )

    # Sort by confidence.
    sortind = [i for (v, i) in sorted((v, i) for (i, v) in enumerate(pred_confs))][::-1]

    # Do the actual matching.
    tp = []  # Accumulator of true positives.
    fp = []  # Accumulator of false positives.
    conf = []  # Accumulator of confidences.

    # match_data holds the extra metrics we calculate for each match.
    match_data = {
        "trans_err": [],
        "vel_err": [],
        "scale_err": [],
        "orient_err": [],
        "attr_err": [],
        "conf": [],
    }

    # ---------------------------------------------
    # Match and accumulate match data.
    # ---------------------------------------------

    taken = set()  # Initially no gt bounding box is matched.
    for ind in sortind:
        pred_box = pred_boxes_list[ind]
        min_dist = np.inf
        match_gt_idx = None

        for gt_idx, gt_box in enumerate(gt_boxes[pred_box.sample_token]):
            # Find closest match among ground truth boxes
            if gt_box.detection_name == class_name and (pred_box.sample_token, gt_idx) not in taken:
                this_distance = dist_fcn(gt_box, pred_box)
                if this_distance < min_dist:
                    min_dist = this_distance
                    match_gt_idx = gt_idx

        # If the closest match is close enough according to threshold we have a match!
        is_match = min_dist < dist_th

        if is_match:
            taken.add((pred_box.sample_token, match_gt_idx))

            #  Update tp, fp and confs.
            tp.append(1)
            fp.append(0)
            conf.append(pred_box.detection_score)

            # Since it is a match, update match data also.
            gt_box_match = gt_boxes[pred_box.sample_token][match_gt_idx]

            match_data["trans_err"].append(center_distance(gt_box_match, pred_box))
            match_data["vel_err"].append(velocity_l2(gt_box_match, pred_box))
            match_data["scale_err"].append(1 - scale_iou(gt_box_match, pred_box))

            # Barrier orientation is only determined up to 180 degree. (For cones orientation is discarded later)
            period = np.pi if class_name == "barrier" else 2 * np.pi
            match_data["orient_err"].append(yaw_diff(gt_box_match, pred_box, period=period))

            match_data["attr_err"].append(1 - attr_acc(gt_box_match, pred_box))
            match_data["conf"].append(pred_box.detection_score)

        else:
            # No match. Mark this as a false positive.
            tp.append(0)
            fp.append(1)
            conf.append(pred_box.detection_score)

    # Check if we have any matches. If not, just return a "no predictions" array.
    if len(match_data["trans_err"]) == 0:
        return DetectionMetricData.no_predictions()

    # ---------------------------------------------
    # Calculate and interpolate precision and recall
    # ---------------------------------------------

    # Accumulate.
    tp = np.cumsum(tp).astype(float)
    fp = np.cumsum(fp).astype(float)
    conf = np.array(conf)

    # Calculate precision and recall.
    prec = tp / (fp + tp)
    rec = tp / float(npos)

    rec_interp = np.linspace(0, 1, DetectionMetricData.nelem)  # 101 steps, from 0% to 100% recall.
    prec = np.interp(rec_interp, rec, prec, right=0)
    conf = np.interp(rec_interp, rec, conf, right=0)
    rec = rec_interp

    # ---------------------------------------------
    # Re-sample the match-data to match, prec, recall and conf.
    # ---------------------------------------------

    for key in match_data.keys():
        if key == "conf":
            continue  # Confidence is used as reference to align with fp and tp. So skip in this step.

        else:
            # For each match_data, we first calculate the accumulated mean.
            tmp = cummean(np.array(match_data[key]))

            # Then interpolate based on the confidences. (Note reversing since np.interp needs increasing arrays)
            match_data[key] = np.interp(conf[::-1], match_data["conf"][::-1], tmp[::-1])[::-1]

    # ---------------------------------------------
    # Done. Instantiate MetricData and return
    # ---------------------------------------------
    return DetectionMetricData(
        recall=rec,
        precision=prec,
        confidence=conf,
        trans_err=match_data["trans_err"],
        vel_err=match_data["vel_err"],
        scale_err=match_data["scale_err"],
        orient_err=match_data["orient_err"],
        attr_err=match_data["attr_err"],
    )


def calc_ap(md: DetectionMetricData, min_recall: float, min_precision: float) -> float:
    """Calculated average precision."""

    assert 0 <= min_precision < 1
    assert 0 <= min_recall <= 1

    prec = np.copy(md.precision)
    prec = prec[
        round(100 * min_recall) + 1 :
    ]  # Clip low recalls. +1 to exclude the min recall bin.
    prec -= min_precision  # Clip low precision
    prec[prec < 0] = 0
    return float(np.mean(prec)) / (1.0 - min_precision)


def calc_tp(md: DetectionMetricData, min_recall: float, metric_name: str) -> float:
    """Calculates true positive errors."""

    first_ind = round(100 * min_recall) + 1  # +1 to exclude the error at min recall.
    last_ind = (
        md.max_recall_ind
    )  # First instance of confidence = 0 is index of max achieved recall.
    if last_ind < first_ind:
        return 1.0  # Assign 1 here. If this happens for all classes, the score for that TP metric will be 0.
    else:
        return float(
            np.mean(getattr(md, metric_name)[first_ind : last_ind + 1])
        )  # +1 to include error at max recall.



"""Utilities for creating point cloud input representation."""
from dataclasses import dataclass
from typing import List, Tuple

import numpy as np

# from constants import DO_CLASSES, DYNAMIC_OBJECTS, SO_CLASSES, STATIC_OBJECTS
# from plot_objects_annot_on_image import ObjectAnnotationHandler, get_annotations_files, read_anno_content


# AVAILABLE_PROJECTS = [DYNAMIC_OBJECTS, STATIC_OBJECTS]
# Dynamic Object classes for BEV visualization
DO_CLASSES = ("Vehicle", "VulnerableVehicle", "Pedestrian", "Animal")
# Static objects classes for BEV visualization
SO_CLASSES = (
    "TrafficSign",
    "PoleObject",
    "TrafficGuide",
    "TrafficBeacon",
    "TrafficSignal",
    "DynamicBarrier",
    "Inconclusive",
)


@dataclass
class BEVSettings:
    """All relevant information about the input."""

    # pylint: disable=too-many-instance-attributes
    # General settings
    grid_min: np.ndarray = np.array([-50.0, 0.0])
    grid_max: np.ndarray = np.array([50.0, 100.0])
    grid_cell_size: float = 0.1  # Default in PIXOR: 0.1

    # Pixor settings
    pixor_z_min: float = -2.4
    pixor_z_max: float = 1.0

    # === Not configurable (automatically calculated) ===
    # General
    grid_channels: int = 0
    grid_res: Tuple[int, int] = (0, 0)
    grid_shape: Tuple[int, int, int] = (0, 0, 0)

    classes: Tuple[str] = DO_CLASSES + SO_CLASSES

    def get_class_name(self, idx: int) -> str:
        """Retrieve class name from index."""
        return self.classes[idx]

    def encode_classes(self, classes_to_encode: List[str]) -> List[int]:
        """Retrieve classes indexes for a list of class names."""
        return [self.classes.index(entry) for entry in classes_to_encode]

    def __post_init__(self):
        """Post initialize fields."""
        # BEV Grid
        grid_res = (self.grid_max - self.grid_min) / self.grid_cell_size
        self.grid_res = tuple(grid_res.astype(int))
        self.grid_channels = int((self.pixor_z_max - self.pixor_z_min) / self.grid_cell_size) + 3
        self.grid_shape = (self.grid_channels,) + self.grid_res


def create_pointcloud_input(points: np.ndarray, settings: BEVSettings) -> np.ndarray:
    """Create input representation from raw data.

    Args:
        points: Point cloud [N, 4] containing ['x', 'y', 'z', 'intensity'].
        settings: Settings defining the input format.

    Returns:
        input_: An instance of the encoded input point cloud.

    """
    # Truncate points according to BEV ranges
    mask = get_grid_mask(points, settings)
    points = points[mask]

    point_indices_xy = get_grid_indices_xy(points[:, :3], settings)

    input_ = _create_pointcloud_input_pixor(points, point_indices_xy, settings)

    return input_


def _create_pointcloud_input_pixor(
    points: np.ndarray, point_indices_xy: np.ndarray, settings: BEVSettings
) -> np.ndarray:
    """Create PIXOR-style input representation.

    Args:
        points: Point cloud [N, 4] containing ['x', 'y', 'z', 'intensity'].
        point_indices_xy: The corresponding xy-indices for the points.
        settings: Settings defining the input format.

    Returns:
        A PIXOR style BEV projection of the input point cloud.

    """
    point_indices_c = np.cast["int32"](
        (points[:, 2] - settings.pixor_z_min) / settings.grid_cell_size
    )
    point_indices_c = 1 + np.clip(
        point_indices_c,
        a_min=-1,
        a_max=settings.grid_channels - 3,
    )
    point_indices_cxy = tuple(
        np.transpose(
            np.concatenate([np.expand_dims(point_indices_c, axis=-1), point_indices_xy], axis=-1)
        ).reshape(3, -1)
    )

    n_points = points.shape[0]

    point_indices_intensity_c = np.repeat(settings.grid_channels - 1, n_points)
    point_indices_intensity_cxy = tuple(
        np.transpose(
            np.concatenate(
                [np.expand_dims(point_indices_intensity_c, axis=-1), point_indices_xy], axis=-1
            )
        ).reshape(3, -1)
    )

    # Define the update per index (currently just occupancy)
    updates = np.ones((n_points,))
    updates_intensity = points[:, 3]

    # Create occupancy grid (with intensity)
    input_ = np.zeros(settings.grid_shape, dtype=np.float32)
    input_[point_indices_cxy] = updates  # Occupancy.
    input_[point_indices_intensity_cxy] = updates_intensity  # Intensity.

    return input_


def get_grid_mask(cloud: np.ndarray, settings: BEVSettings) -> np.ndarray:
    """Get the boolean mask to filter out points within the grid.

    Args:
        cloud: The input point cloud [N, 3]
        settings: The input definition.

    Returns:
        mask: Boolean mask containing true for points within grid.

    """
    scaled_xy = cloud[:, :2] - settings.grid_min
    bev_max = settings.grid_max - settings.grid_min
    mask = np.all(((scaled_xy >= 0) & (scaled_xy < bev_max)), axis=-1)
    return mask


def get_grid_indices_xy(cloud: np.ndarray, settings: BEVSettings) -> np.ndarray:
    """Get grid indices from point cloud.

    Args:
        cloud: Point cloud of x,y,z coordinates [N, 3].
        settings: Settings defining the grid size and resolution.

    Returns:
        indices_xy: The xy grid indices of each point [N, 2].

    """
    # Convert points to indices
    indices_xy = np.cast["int32"]((cloud[:, :2] - settings.grid_min) / settings.grid_cell_size)
    return indices_xy


def filter_point_cloud(cloud: np.ndarray, angle: np.ndarray, cam_pos: np.ndarray) -> np.ndarray:
    """Filter out points outside of camera-centered frustum."""
    cloud_xy = cloud[:, :2] - cam_pos[:2]
    point_angles = np.arctan2(cloud_xy[:, 1], cloud_xy[:, 0])
    mask = np.logical_or(point_angles < angle[0], point_angles > angle[1])
    return cloud[mask]


def get_objects_for_bev(
    seq_folder: str,
    annotation_projects: List[str],
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Get annotation objects for bird eye view visualization.

    Args:
        seq_folder: path to the sequence folder.
        annotation_projects: list of projects you want to visualize.
                             Possible projects: DYNAMIC_OBJECTS, STATIC_OBJECTS.

    Returns:
        extracted_anno_objects: positions, dimensions, rotations and classes of bounding boxes
                                to be visualized.

    """
    anno_project_files = get_annotations_files(seq_folder)
    anno_objects = []
    for proj in annotation_projects:
        if proj in AVAILABLE_PROJECTS:
            anno_file = anno_project_files[proj]
            anno_content = read_anno_content(anno_file)
            anno_objects.extend(list(ObjectAnnotationHandler.from_annotations(anno_content)))
        else:
            raise Exception(
                f"Project {proj} is not available to plot. "
                f"Available projects: {*AVAILABLE_PROJECTS,}."
            )
    positions = [obj[2].marking3d.get("Location")[:2] for obj in anno_objects if obj[2].marking3d]
    dimensions = [obj[2].marking3d.get("Size")[:2] for obj in anno_objects if obj[2].marking3d]
    rotations = [obj[2].marking3d.get("Rotation") for obj in anno_objects if obj[2].marking3d]
    classes = [[obj[0]] for obj in anno_objects if obj[2].marking3d]
    extracted_anno_objects = (
        np.array(classes),
        np.array(positions),
        np.array(dimensions),
        np.array(rotations),
    )
    return extracted_anno_objects



"""Class map for the traffic sign classification dataset."""

CLASS_NAME_TO_CLASS_IDX = {
    "NotListed": 0,
    "Indication_CameraSurveillance": 1,
    "Mandatory_PassOnEitherSide": 2,
    "Mandatory_PassOnThisSideLeft": 3,
    "Mandatory_PassOnThisSideRight": 4,
    "Mandatory_ProceedStraight": 5,
    "Mandatory_ProceedStraightOrTurnRight": 6,
    "Mandatory_ProceedStraightOrTurnLeft": 7,
    "Mandatory_Roundabout": 8,
    "Mandatory_TurnAhead": 9,
    "Mandatory_TurnLeft": 10,
    "Mandatory_TurnLeftAhead": 11,
    "Mandatory_TurnRight": 12,
    "Mandatory_TurnRightAhead": 13,
    "Priority_GiveWayOncoming": 14,
    "Priority_GiveWay": 15,
    "Priority_PriorityRoadBegin": 16,
    "Priority_PriorityRoadEnd": 17,
    "Priority_PrioOverOncoming": 18,
    "Priority_Stop": 19,
    "Prohibitory_MaximumSpeedLimit5End": 20,
    "Prohibitory_MaximumSpeedLimit5Begin": 21,
    "Prohibitory_MaximumSpeedLimit10End": 22,
    "Prohibitory_MaximumSpeedLimit10Begin": 23,
    "Prohibitory_MaximumSpeedLimit15End": 24,
    "Prohibitory_MaximumSpeedLimit15Begin": 25,
    "Prohibitory_MaximumSpeedLimit20Begin": 26,
    "Prohibitory_MaximumSpeedLimit20End": 27,
    "Prohibitory_MaximumSpeedLimit25End": 28,
    "Prohibitory_MaximumSpeedLimit25Begin": 29,
    "Prohibitory_MaximumSpeedLimit30End": 30,
    "Prohibitory_MaximumSpeedLimit30Begin": 31,
    "Prohibitory_MaximumSpeedLimit35End": 32,
    "Prohibitory_MaximumSpeedLimit35Begin": 33,
    "Prohibitory_MaximumSpeedLimit40End": 34,
    "Prohibitory_MaximumSpeedLimit40Begin": 35,
    "Prohibitory_MaximumSpeedLimit45End": 36,
    "Prohibitory_MaximumSpeedLimit45Begin": 37,
    "Prohibitory_MaximumSpeedLimit50Begin": 38,
    "Prohibitory_MaximumSpeedLimit50End": 39,
    "Prohibitory_MaximumSpeedLimit55Begin": 40,
    "Prohibitory_MaximumSpeedLimit55End": 41,
    "Prohibitory_MaximumSpeedLimit60Begin": 42,
    "Prohibitory_MaximumSpeedLimit60End": 43,
    "Prohibitory_MaximumSpeedLimit65Begin": 44,
    "Prohibitory_MaximumSpeedLimit65End": 45,
    "Prohibitory_MaximumSpeedLimit70Begin": 46,
    "Prohibitory_MaximumSpeedLimit70End": 47,
    "Prohibitory_MaximumSpeedLimit75End": 48,
    "Prohibitory_MaximumSpeedLimit75Begin": 49,
    "Prohibitory_MaximumSpeedLimit80Begin": 50,
    "Prohibitory_MaximumSpeedLimit80End": 51,
    "Prohibitory_MaximumSpeedLimit85Begin": 52,
    "Prohibitory_MaximumSpeedLimit85End": 53,
    "Prohibitory_MaximumSpeedLimit90Begin": 54,
    "Prohibitory_MaximumSpeedLimit90End": 55,
    "Prohibitory_MaximumSpeedLimit95Begin": 56,
    "Prohibitory_MaximumSpeedLimit95End": 57,
    "Prohibitory_MaximumSpeedLimit100End": 58,
    "Prohibitory_MaximumSpeedLimit100Begin": 59,
    "Prohibitory_MaximumSpeedLimit105End": 60,
    "Prohibitory_MaximumSpeedLimit105Begin": 61,
    "Prohibitory_MaximumSpeedLimit110End": 62,
    "Prohibitory_MaximumSpeedLimit110Begin": 63,
    "Prohibitory_MaximumSpeedLimit115Begin": 64,
    "Prohibitory_MaximumSpeedLimit115End": 65,
    "Prohibitory_MaximumSpeedLimit120End": 66,
    "Prohibitory_MaximumSpeedLimit120Begin": 67,
    "Prohibitory_MaximumSpeedLimit125Begin": 68,
    "Prohibitory_MaximumSpeedLimit125End": 69,
    "Prohibitory_MaximumSpeedLimit130End": 70,
    "Prohibitory_MaximumSpeedLimit130Begin": 71,
    "Prohibitory_NoEntry": 72,
    "Prohibitory_NoOvertakingBegin": 73,
    "Prohibitory_NoOvertakingEnd": 74,
    "Prohibitory_NoParking": 75,
    "Prohibitory_NoStopping": 76,
    "Prohibitory_NoTurn": 77,
    "Prohibitory_NoUTurn": 78,
    "Prohibitory_RoadClosed": 79,
    "Prohibitory_SpeedLimitZone5End": 80,
    "Prohibitory_SpeedLimitZone5Begin": 81,
    "Prohibitory_SpeedLimitZone10Begin": 82,
    "Prohibitory_SpeedLimitZone10End": 83,
    "Prohibitory_SpeedLimitZone15End": 84,
    "Prohibitory_SpeedLimitZone15Begin": 85,
    "Prohibitory_SpeedLimitZone20Begin": 86,
    "Prohibitory_SpeedLimitZone20End": 87,
    "Prohibitory_SpeedLimitZone25End": 88,
    "Prohibitory_SpeedLimitZone25Begin": 89,
    "Prohibitory_SpeedLimitZone30End": 90,
    "Prohibitory_SpeedLimitZone30Begin": 91,
    "Prohibitory_SpeedLimitZone35End": 92,
    "Prohibitory_SpeedLimitZone35Begin": 93,
    "Prohibitory_SpeedLimitZone40End": 94,
    "Prohibitory_SpeedLimitZone40Begin": 95,
    "Prohibitory_SpeedLimitZone45Begin": 96,
    "Prohibitory_SpeedLimitZone45End": 97,
    "Prohibitory_SpeedLimitZone50Begin": 98,
    "Prohibitory_SpeedLimitZone50End": 99,
    "Prohibitory_SpeedLimitZone55End": 100,
    "Prohibitory_SpeedLimitZone55Begin": 101,
    "Prohibitory_SpeedLimitZone60End": 102,
    "Prohibitory_SpeedLimitZone60Begin": 103,
    "Prohibitory_SpeedLimitZone65End": 104,
    "Prohibitory_SpeedLimitZone65Begin": 105,
    "Prohibitory_SpeedLimitZone70End": 106,
    "Prohibitory_SpeedLimitZone70Begin": 107,
    "Prohibitory_SpeedLimitZone75End": 108,
    "Prohibitory_SpeedLimitZone75Begin": 109,
    "Prohibitory_SpeedLimitZone80Begin": 110,
    "Prohibitory_SpeedLimitZone80End": 111,
    "Prohibitory_SpeedLimitZone85End": 112,
    "Prohibitory_SpeedLimitZone85Begin": 113,
    "Prohibitory_SpeedLimitZone90Begin": 114,
    "Prohibitory_SpeedLimitZone90End": 115,
    "Prohibitory_SpeedLimitZone95Begin": 116,
    "Prohibitory_SpeedLimitZone95End": 117,
    "Prohibitory_SpeedLimitZone100End": 118,
    "Prohibitory_SpeedLimitZone100Begin": 119,
    "Prohibitory_SpeedLimitZone105End": 120,
    "Prohibitory_SpeedLimitZone105Begin": 121,
    "Prohibitory_SpeedLimitZone110End": 122,
    "Prohibitory_SpeedLimitZone110Begin": 123,
    "Prohibitory_SpeedLimitZone115Begin": 124,
    "Prohibitory_SpeedLimitZone115End": 125,
    "Prohibitory_SpeedLimitZone120Begin": 126,
    "Prohibitory_SpeedLimitZone120End": 127,
    "Prohibitory_SpeedLimitZone125Begin": 128,
    "Prohibitory_SpeedLimitZone125End": 129,
    "Prohibitory_SpeedLimitZone130End": 130,
    "Prohibitory_SpeedLimitZone130Begin": 131,
    "RoadType_MotorwayBegin": 132,
    "RoadType_MotorwayEnd": 133,
    "Special_VulnurableRoadUserCrossing": 134,
    "Special_VulnurableRoadUserPathWay": 135,
    "Warning_Animal": 136,
    "Warning_Children": 137,
    "Warning_CongestionAhead": 138,
    "Warning_Crossroads": 139,
    "Warning_Crossing": 140,
    "Warning_Curve": 141,
    "Warning_Cyclists": 142,
    "Warning_DoubleCurve": 143,
    "Warning_GenericWarning": 144,
    "Warning_MergingTraffic": 145,
    "Warning_RoadBump": 146,
    "Warning_RoadWorkEnd": 147,
    "Warning_RoadWorkBegin": 148,
    "Warning_RoadNarrows": 149,
    "Warning_RoughRoad": 150,
    "Warning_Roundabout": 151,
    "Warning_Slippery": 152,
    "Warning_TrafficSignalAhead": 153,
    "Warning_TunnelAhead": 154,
    "Warning_TwoWayTraffic": 155,
}

CLASS_IDX_TO_CLASS_NAME = {v: k for k, v in CLASS_NAME_TO_CLASS_IDX.items()}

TRAIN_SAMPLES_PER_CLASS_NAME = {
    "NotListed": 211898,
    "Indication_CameraSurveillance": 44,
    "Mandatory_PassOnEitherSide": 1003,
    "Mandatory_PassOnThisSideLeft": 1052,
    "Mandatory_PassOnThisSideRight": 15819,
    "Mandatory_ProceedStraight": 6892,
    "Mandatory_ProceedStraightOrTurnRight": 1805,
    "Mandatory_ProceedStraightOrTurnLeft": 548,
    "Mandatory_Roundabout": 1908,
    "Mandatory_TurnAhead": 145,
    "Mandatory_TurnLeft": 386,
    "Mandatory_TurnLeftAhead": 1047,
    "Mandatory_TurnRight": 811,
    "Mandatory_TurnRightAhead": 2087,
    "Priority_GiveWayOncoming": 26,
    "Priority_GiveWay": 17132,
    "Priority_PriorityRoadBegin": 20270,
    "Priority_PriorityRoadEnd": 1846,
    "Priority_PrioOverOncoming": 34,
    "Priority_Stop": 1463,
    "Prohibitory_MaximumSpeedLimit5End": 0,
    "Prohibitory_MaximumSpeedLimit5Begin": 49,
    "Prohibitory_MaximumSpeedLimit10End": 0,
    "Prohibitory_MaximumSpeedLimit10Begin": 156,
    "Prohibitory_MaximumSpeedLimit15End": 0,
    "Prohibitory_MaximumSpeedLimit15Begin": 11,
    "Prohibitory_MaximumSpeedLimit20Begin": 155,
    "Prohibitory_MaximumSpeedLimit20End": 2,
    "Prohibitory_MaximumSpeedLimit25End": 0,
    "Prohibitory_MaximumSpeedLimit25Begin": 6,
    "Prohibitory_MaximumSpeedLimit30End": 118,
    "Prohibitory_MaximumSpeedLimit30Begin": 3141,
    "Prohibitory_MaximumSpeedLimit35End": 0,
    "Prohibitory_MaximumSpeedLimit35Begin": 2,
    "Prohibitory_MaximumSpeedLimit40End": 35,
    "Prohibitory_MaximumSpeedLimit40Begin": 1524,
    "Prohibitory_MaximumSpeedLimit45End": 0,
    "Prohibitory_MaximumSpeedLimit45Begin": 1,
    "Prohibitory_MaximumSpeedLimit50Begin": 1624,
    "Prohibitory_MaximumSpeedLimit50End": 16,
    "Prohibitory_MaximumSpeedLimit55Begin": 0,
    "Prohibitory_MaximumSpeedLimit55End": 0,
    "Prohibitory_MaximumSpeedLimit60Begin": 1260,
    "Prohibitory_MaximumSpeedLimit60End": 18,
    "Prohibitory_MaximumSpeedLimit65Begin": 2,
    "Prohibitory_MaximumSpeedLimit65End": 0,
    "Prohibitory_MaximumSpeedLimit70Begin": 2887,
    "Prohibitory_MaximumSpeedLimit70End": 48,
    "Prohibitory_MaximumSpeedLimit75End": 0,
    "Prohibitory_MaximumSpeedLimit75Begin": 0,
    "Prohibitory_MaximumSpeedLimit80Begin": 806,
    "Prohibitory_MaximumSpeedLimit80End": 25,
    "Prohibitory_MaximumSpeedLimit85Begin": 1,
    "Prohibitory_MaximumSpeedLimit85End": 0,
    "Prohibitory_MaximumSpeedLimit90Begin": 627,
    "Prohibitory_MaximumSpeedLimit90End": 10,
    "Prohibitory_MaximumSpeedLimit95Begin": 1,
    "Prohibitory_MaximumSpeedLimit95End": 0,
    "Prohibitory_MaximumSpeedLimit100End": 12,
    "Prohibitory_MaximumSpeedLimit100Begin": 232,
    "Prohibitory_MaximumSpeedLimit105End": 0,
    "Prohibitory_MaximumSpeedLimit105Begin": 0,
    "Prohibitory_MaximumSpeedLimit110End": 3,
    "Prohibitory_MaximumSpeedLimit110Begin": 120,
    "Prohibitory_MaximumSpeedLimit115Begin": 0,
    "Prohibitory_MaximumSpeedLimit115End": 0,
    "Prohibitory_MaximumSpeedLimit120End": 9,
    "Prohibitory_MaximumSpeedLimit120Begin": 63,
    "Prohibitory_MaximumSpeedLimit125Begin": 0,
    "Prohibitory_MaximumSpeedLimit125End": 0,
    "Prohibitory_MaximumSpeedLimit130End": 2,
    "Prohibitory_MaximumSpeedLimit130Begin": 19,
    "Prohibitory_NoEntry": 18597,
    "Prohibitory_NoOvertakingBegin": 1663,
    "Prohibitory_NoOvertakingEnd": 93,
    "Prohibitory_NoParking": 6574,
    "Prohibitory_NoStopping": 20033,
    "Prohibitory_NoTurn": 3956,
    "Prohibitory_NoUTurn": 1108,
    "Prohibitory_RoadClosed": 2051,
    "Prohibitory_SpeedLimitZone5End": 0,
    "Prohibitory_SpeedLimitZone5Begin": 0,
    "Prohibitory_SpeedLimitZone10Begin": 19,
    "Prohibitory_SpeedLimitZone10End": 1,
    "Prohibitory_SpeedLimitZone15End": 0,
    "Prohibitory_SpeedLimitZone15Begin": 1,
    "Prohibitory_SpeedLimitZone20Begin": 90,
    "Prohibitory_SpeedLimitZone20End": 4,
    "Prohibitory_SpeedLimitZone25End": 0,
    "Prohibitory_SpeedLimitZone25Begin": 0,
    "Prohibitory_SpeedLimitZone30End": 235,
    "Prohibitory_SpeedLimitZone30Begin": 1784,
    "Prohibitory_SpeedLimitZone35End": 0,
    "Prohibitory_SpeedLimitZone35Begin": 1,
    "Prohibitory_SpeedLimitZone40End": 2,
    "Prohibitory_SpeedLimitZone40Begin": 51,
    "Prohibitory_SpeedLimitZone45Begin": 0,
    "Prohibitory_SpeedLimitZone45End": 0,
    "Prohibitory_SpeedLimitZone50Begin": 424,
    "Prohibitory_SpeedLimitZone50End": 109,
    "Prohibitory_SpeedLimitZone55End": 0,
    "Prohibitory_SpeedLimitZone55Begin": 0,
    "Prohibitory_SpeedLimitZone60End": 0,
    "Prohibitory_SpeedLimitZone60Begin": 27,
    "Prohibitory_SpeedLimitZone65End": 0,
    "Prohibitory_SpeedLimitZone65Begin": 0,
    "Prohibitory_SpeedLimitZone70End": 0,
    "Prohibitory_SpeedLimitZone70Begin": 7,
    "Prohibitory_SpeedLimitZone75End": 0,
    "Prohibitory_SpeedLimitZone75Begin": 1,
    "Prohibitory_SpeedLimitZone80Begin": 9,
    "Prohibitory_SpeedLimitZone80End": 0,
    "Prohibitory_SpeedLimitZone85End": 0,
    "Prohibitory_SpeedLimitZone85Begin": 1,
    "Prohibitory_SpeedLimitZone90Begin": 2,
    "Prohibitory_SpeedLimitZone90End": 0,
    "Prohibitory_SpeedLimitZone95Begin": 0,
    "Prohibitory_SpeedLimitZone95End": 0,
    "Prohibitory_SpeedLimitZone100End": 0,
    "Prohibitory_SpeedLimitZone100Begin": 5,
    "Prohibitory_SpeedLimitZone105End": 0,
    "Prohibitory_SpeedLimitZone105Begin": 0,
    "Prohibitory_SpeedLimitZone110End": 0,
    "Prohibitory_SpeedLimitZone110Begin": 0,
    "Prohibitory_SpeedLimitZone115Begin": 0,
    "Prohibitory_SpeedLimitZone115End": 0,
    "Prohibitory_SpeedLimitZone120Begin": 4,
    "Prohibitory_SpeedLimitZone120End": 0,
    "Prohibitory_SpeedLimitZone125Begin": 0,
    "Prohibitory_SpeedLimitZone125End": 0,
    "Prohibitory_SpeedLimitZone130End": 0,
    "Prohibitory_SpeedLimitZone130Begin": 0,
    "RoadType_MotorwayBegin": 161,
    "RoadType_MotorwayEnd": 100,
    "Special_VulnurableRoadUserCrossing": 31023,
    "Special_VulnurableRoadUserPathWay": 6673,
    "Warning_Animal": 294,
    "Warning_Children": 621,
    "Warning_CongestionAhead": 109,
    "Warning_Crossroads": 1061,
    "Warning_Crossing": 751,
    "Warning_Curve": 392,
    "Warning_Cyclists": 398,
    "Warning_DoubleCurve": 92,
    "Warning_GenericWarning": 2097,
    "Warning_MergingTraffic": 323,
    "Warning_RoadBump": 620,
    "Warning_RoadWorkEnd": 7,
    "Warning_RoadWorkBegin": 2485,
    "Warning_RoadNarrows": 841,
    "Warning_RoughRoad": 230,
    "Warning_Roundabout": 205,
    "Warning_Slippery": 149,
    "Warning_TrafficSignalAhead": 409,
    "Warning_TunnelAhead": 4,
    "Warning_TwoWayTraffic": 234,
}

VAL_SAMPLES_PER_CLASS_NAME = {
    "NotListed": 21851,
    "Indication_CameraSurveillance": 8,
    "Mandatory_PassOnEitherSide": 83,
    "Mandatory_PassOnThisSideLeft": 107,
    "Mandatory_PassOnThisSideRight": 1775,
    "Mandatory_ProceedStraight": 669,
    "Mandatory_ProceedStraightOrTurnRight": 147,
    "Mandatory_ProceedStraightOrTurnLeft": 62,
    "Mandatory_Roundabout": 183,
    "Mandatory_TurnAhead": 9,
    "Mandatory_TurnLeft": 26,
    "Mandatory_TurnLeftAhead": 127,
    "Mandatory_TurnRight": 94,
    "Mandatory_TurnRightAhead": 197,
    "Priority_GiveWayOncoming": 3,
    "Priority_GiveWay": 1798,
    "Priority_PriorityRoadBegin": 2026,
    "Priority_PriorityRoadEnd": 108,
    "Priority_PrioOverOncoming": 7,
    "Priority_Stop": 158,
    "Prohibitory_MaximumSpeedLimit5End": 0,
    "Prohibitory_MaximumSpeedLimit5Begin": 1,
    "Prohibitory_MaximumSpeedLimit10End": 0,
    "Prohibitory_MaximumSpeedLimit10Begin": 10,
    "Prohibitory_MaximumSpeedLimit15End": 0,
    "Prohibitory_MaximumSpeedLimit15Begin": 4,
    "Prohibitory_MaximumSpeedLimit20Begin": 19,
    "Prohibitory_MaximumSpeedLimit20End": 0,
    "Prohibitory_MaximumSpeedLimit25End": 0,
    "Prohibitory_MaximumSpeedLimit25Begin": 2,
    "Prohibitory_MaximumSpeedLimit30End": 11,
    "Prohibitory_MaximumSpeedLimit30Begin": 444,
    "Prohibitory_MaximumSpeedLimit35End": 0,
    "Prohibitory_MaximumSpeedLimit35Begin": 0,
    "Prohibitory_MaximumSpeedLimit40End": 0,
    "Prohibitory_MaximumSpeedLimit40Begin": 89,
    "Prohibitory_MaximumSpeedLimit45End": 0,
    "Prohibitory_MaximumSpeedLimit45Begin": 0,
    "Prohibitory_MaximumSpeedLimit50Begin": 192,
    "Prohibitory_MaximumSpeedLimit50End": 3,
    "Prohibitory_MaximumSpeedLimit55Begin": 0,
    "Prohibitory_MaximumSpeedLimit55End": 0,
    "Prohibitory_MaximumSpeedLimit60Begin": 89,
    "Prohibitory_MaximumSpeedLimit60End": 5,
    "Prohibitory_MaximumSpeedLimit65Begin": 0,
    "Prohibitory_MaximumSpeedLimit65End": 0,
    "Prohibitory_MaximumSpeedLimit70Begin": 276,
    "Prohibitory_MaximumSpeedLimit70End": 1,
    "Prohibitory_MaximumSpeedLimit75End": 0,
    "Prohibitory_MaximumSpeedLimit75Begin": 0,
    "Prohibitory_MaximumSpeedLimit80Begin": 69,
    "Prohibitory_MaximumSpeedLimit80End": 1,
    "Prohibitory_MaximumSpeedLimit85Begin": 0,
    "Prohibitory_MaximumSpeedLimit85End": 0,
    "Prohibitory_MaximumSpeedLimit90Begin": 67,
    "Prohibitory_MaximumSpeedLimit90End": 1,
    "Prohibitory_MaximumSpeedLimit95Begin": 0,
    "Prohibitory_MaximumSpeedLimit95End": 0,
    "Prohibitory_MaximumSpeedLimit100End": 1,
    "Prohibitory_MaximumSpeedLimit100Begin": 21,
    "Prohibitory_MaximumSpeedLimit105End": 0,
    "Prohibitory_MaximumSpeedLimit105Begin": 0,
    "Prohibitory_MaximumSpeedLimit110End": 0,
    "Prohibitory_MaximumSpeedLimit110Begin": 27,
    "Prohibitory_MaximumSpeedLimit115Begin": 0,
    "Prohibitory_MaximumSpeedLimit115End": 0,
    "Prohibitory_MaximumSpeedLimit120End": 2,
    "Prohibitory_MaximumSpeedLimit120Begin": 6,
    "Prohibitory_MaximumSpeedLimit125Begin": 0,
    "Prohibitory_MaximumSpeedLimit125End": 0,
    "Prohibitory_MaximumSpeedLimit130End": 0,
    "Prohibitory_MaximumSpeedLimit130Begin": 12,
    "Prohibitory_NoEntry": 1771,
    "Prohibitory_NoOvertakingBegin": 249,
    "Prohibitory_NoOvertakingEnd": 3,
    "Prohibitory_NoParking": 777,
    "Prohibitory_NoStopping": 1798,
    "Prohibitory_NoTurn": 379,
    "Prohibitory_NoUTurn": 94,
    "Prohibitory_RoadClosed": 158,
    "Prohibitory_SpeedLimitZone5End": 0,
    "Prohibitory_SpeedLimitZone5Begin": 0,
    "Prohibitory_SpeedLimitZone10Begin": 2,
    "Prohibitory_SpeedLimitZone10End": 0,
    "Prohibitory_SpeedLimitZone15End": 0,
    "Prohibitory_SpeedLimitZone15Begin": 0,
    "Prohibitory_SpeedLimitZone20Begin": 17,
    "Prohibitory_SpeedLimitZone20End": 1,
    "Prohibitory_SpeedLimitZone25End": 0,
    "Prohibitory_SpeedLimitZone25Begin": 0,
    "Prohibitory_SpeedLimitZone30End": 32,
    "Prohibitory_SpeedLimitZone30Begin": 173,
    "Prohibitory_SpeedLimitZone35End": 0,
    "Prohibitory_SpeedLimitZone35Begin": 0,
    "Prohibitory_SpeedLimitZone40End": 1,
    "Prohibitory_SpeedLimitZone40Begin": 6,
    "Prohibitory_SpeedLimitZone45Begin": 0,
    "Prohibitory_SpeedLimitZone45End": 0,
    "Prohibitory_SpeedLimitZone50Begin": 39,
    "Prohibitory_SpeedLimitZone50End": 9,
    "Prohibitory_SpeedLimitZone55End": 0,
    "Prohibitory_SpeedLimitZone55Begin": 0,
    "Prohibitory_SpeedLimitZone60End": 0,
    "Prohibitory_SpeedLimitZone60Begin": 0,
    "Prohibitory_SpeedLimitZone65End": 0,
    "Prohibitory_SpeedLimitZone65Begin": 0,
    "Prohibitory_SpeedLimitZone70End": 0,
    "Prohibitory_SpeedLimitZone70Begin": 0,
    "Prohibitory_SpeedLimitZone75End": 0,
    "Prohibitory_SpeedLimitZone75Begin": 0,
    "Prohibitory_SpeedLimitZone80Begin": 0,
    "Prohibitory_SpeedLimitZone80End": 0,
    "Prohibitory_SpeedLimitZone85End": 0,
    "Prohibitory_SpeedLimitZone85Begin": 0,
    "Prohibitory_SpeedLimitZone90Begin": 0,
    "Prohibitory_SpeedLimitZone90End": 0,
    "Prohibitory_SpeedLimitZone95Begin": 0,
    "Prohibitory_SpeedLimitZone95End": 0,
    "Prohibitory_SpeedLimitZone100End": 0,
    "Prohibitory_SpeedLimitZone100Begin": 0,
    "Prohibitory_SpeedLimitZone105End": 0,
    "Prohibitory_SpeedLimitZone105Begin": 0,
    "Prohibitory_SpeedLimitZone110End": 0,
    "Prohibitory_SpeedLimitZone110Begin": 0,
    "Prohibitory_SpeedLimitZone115Begin": 0,
    "Prohibitory_SpeedLimitZone115End": 0,
    "Prohibitory_SpeedLimitZone120Begin": 0,
    "Prohibitory_SpeedLimitZone120End": 0,
    "Prohibitory_SpeedLimitZone125Begin": 0,
    "Prohibitory_SpeedLimitZone125End": 0,
    "Prohibitory_SpeedLimitZone130End": 0,
    "Prohibitory_SpeedLimitZone130Begin": 0,
    "RoadType_MotorwayBegin": 32,
    "RoadType_MotorwayEnd": 11,
    "Special_VulnurableRoadUserCrossing": 2991,
    "Special_VulnurableRoadUserPathWay": 681,
    "Warning_Animal": 32,
    "Warning_Children": 91,
    "Warning_CongestionAhead": 12,
    "Warning_Crossroads": 173,
    "Warning_Crossing": 78,
    "Warning_Curve": 50,
    "Warning_Cyclists": 37,
    "Warning_DoubleCurve": 17,
    "Warning_GenericWarning": 164,
    "Warning_MergingTraffic": 32,
    "Warning_RoadBump": 78,
    "Warning_RoadWorkEnd": 0,
    "Warning_RoadWorkBegin": 299,
    "Warning_RoadNarrows": 81,
    "Warning_RoughRoad": 34,
    "Warning_Roundabout": 16,
    "Warning_Slippery": 26,
    "Warning_TrafficSignalAhead": 31,
    "Warning_TunnelAhead": 1,
    "Warning_TwoWayTraffic": 41,
}

N_VAL_SAMPLES_LOWER_BOUND = 10

CLASSES_USED_FOR_VALIDATION = [
    name for name, count in VAL_SAMPLES_PER_CLASS_NAME.items() if count >= N_VAL_SAMPLES_LOWER_BOUND
]


def get_class_name(class_idx: int) -> str:
    """Get class name from class index."""
    if class_idx not in CLASS_IDX_TO_CLASS_NAME:
        raise ValueError(f"Unknown class index: {class_idx}")
    return CLASS_IDX_TO_CLASS_NAME.get(class_idx)


def get_class_idx(class_name: str) -> int:
    """Get class index from class name."""
    if class_name not in CLASS_NAME_TO_CLASS_IDX:
        raise ValueError(f"Unknown class name: {class_name}")
    return CLASS_NAME_TO_CLASS_IDX.get(class_name)



from .evaluate import nuscenes_evaluate as nuscenes_evaluate
from .nuscenes_eval.common.data_classes import EvalBoxes as EvalBoxes
from .nuscenes_eval.detection.data_classes import DetectionBox as DetectionBox



import cv2
import numpy as np


def _create_image_mask(img_shape):
    """
    Creating empty (black) image
    """
    return np.zeros(img_shape, dtype=np.uint8)


def polygon_to_array(polygon):
    """
    Reshape and make sure the polygon is numpy
    """
    poly_array = np.array(polygon).reshape((-1, 1, 2)).astype(np.int32)
    return poly_array


def polygon_to_mask(polygon, img_shape=(2168, 3848, 3), fill_color=(100, 0, 0)):
    """
    Create mask with color for a single polygon
    """
    poly_array = polygon_to_array(polygon)
    image_mask = _create_image_mask(img_shape)
    cv2.fillPoly(image_mask, [poly_array], fill_color)
    return image_mask


def polygons_to_binary_mask(polygons, img_shape=(2168, 3848)):
    """
    Create binary mask for list of polygons
    """
    assert len(img_shape) == 2

    mask = _create_image_mask(img_shape)
    for polygon in polygons:
        mask += polygon_to_mask(polygon, img_shape, 1)
    return mask.astype(bool)



# nuScenes dev-kit.
# Code written by Oscar Beijbom and Varun Bankiti, 2019.

DETECTION_NAMES = [
    "Vehicle",
    "VulnerableVehicle",
    "Pedestrian",
    "TrafficSign",
    "TrafficSignal",
]

PRETTY_DETECTION_NAMES = {
    "car": "Car",
    "truck": "Truck",
    "bus": "Bus",
    "trailer": "Trailer",
    "construction_vehicle": "Constr. Veh.",
    "pedestrian": "Pedestrian",
    "motorcycle": "Motorcycle",
    "bicycle": "Bicycle",
    "traffic_cone": "Traffic Cone",
    "barrier": "Barrier",
}

DETECTION_COLORS = {
    "car": "C0",
    "truck": "C1",
    "bus": "C2",
    "trailer": "C3",
    "construction_vehicle": "C4",
    "pedestrian": "C5",
    "motorcycle": "C6",
    "bicycle": "C7",
    "traffic_cone": "C8",
    "barrier": "C9",
}

ATTRIBUTE_NAMES = [
    "pedestrian.moving",
    "pedestrian.sitting_lying_down",
    "pedestrian.standing",
    "cycle.with_rider",
    "cycle.without_rider",
    "vehicle.moving",
    "vehicle.parked",
    "vehicle.stopped",
]

PRETTY_ATTRIBUTE_NAMES = {
    "pedestrian.moving": "Ped. Moving",
    "pedestrian.sitting_lying_down": "Ped. Sitting",
    "pedestrian.standing": "Ped. Standing",
    "cycle.with_rider": "Cycle w/ Rider",
    "cycle.without_rider": "Cycle w/o Rider",
    "vehicle.moving": "Veh. Moving",
    "vehicle.parked": "Veh. Parked",
    "vehicle.stopped": "Veh. Stopped",
}

TP_METRICS = ["trans_err", "scale_err", "orient_err"]

PRETTY_TP_METRICS = {
    "trans_err": "Trans.",
    "scale_err": "Scale",
    "orient_err": "Orient.",
}

TP_METRICS_UNITS = {
    "trans_err": "m",
    "scale_err": "1-IOU",
    "orient_err": "rad.",
}



from typing import Any, List, Optional

from zod.constants import AnnotationProject, Lidar
from zod.utils.compensation import motion_compensate_scanwise

from .calibration import Calibration
from .info import Information
from .metadata import SequenceMetadata
from .oxts import EgoMotion
from .sensor import LidarData


class ZodSequence:
    def __init__(self, info: Information):
        self.info: Information = info  # holds all the paths to the files
        self._ego_motion: EgoMotion = None  # this is the light-weight version of oxts
        self._oxts: EgoMotion = None
        self._calibration: Calibration = None
        self._metadata: SequenceMetadata = None

    @property
    def ego_motion(self) -> EgoMotion:
        """Get the oxts file."""
        if self._ego_motion is None:
            self._ego_motion = EgoMotion.from_json_path(self.info.ego_motion_path)
        return self._ego_motion

    @property
    def oxts(self) -> EgoMotion:
        """Get the oxts."""
        if self._oxts is None:
            self._oxts = EgoMotion.from_oxts_path(self.info.oxts_path)
        return self._oxts

    @property
    def calibration(self) -> Calibration:
        """Get the calibration."""
        if self._calibration is None:
            self._calibration = Calibration.from_json_path(self.info.calibration_path)
        return self._calibration

    @property
    def metadata(self) -> SequenceMetadata:
        """Get the metadata."""
        if self._metadata is None:
            self._metadata = SequenceMetadata.from_json_path(self.info.metadata_path)
        return self._metadata

    def get_annotation(self, project: AnnotationProject) -> List[Any]:
        """Get the annotation for a given project."""
        anno_frame = self.info.get_key_annotation_frame(project)
        return anno_frame and anno_frame.read()  # read if not None

    def get_lidar(self, start: int = 0, end: int = -1) -> List[LidarData]:
        """Get the point clouds."""
        return [
            lidar_frame.read()
            for lidar_frame in self.info.get_lidar_frames(Lidar.VELODYNE)[start:end]
        ]

    def get_aggregated_lidar(
        self, start: int = 0, end: int = -1, timestamp: Optional[float] = None
    ) -> LidarData:
        """Get the aggregated point cloud."""
        lidar_scans = self.get_lidar(start, end)
        if timestamp is None:
            timestamp = lidar_scans[len(lidar_scans) // 2].core_timestamp
        for i, scan in enumerate(lidar_scans):
            if scan.core_timestamp == timestamp:
                continue
            compensated_scan = motion_compensate_scanwise(
                scan, self.ego_motion, self.calibration.lidars[Lidar.VELODYNE], timestamp
            )
            lidar_scans[i] = compensated_scan
        aggregated = lidar_scans[0]
        aggregated.extend(*lidar_scans[1:])
        return aggregated



# nuScenes dev-kit.
# Code written by Oscar Beijbom, 2019.

from collections import defaultdict
from typing import Dict, List, Tuple

import numpy as np

from ..common.data_classes import EvalBox, MetricData
from ..common.utils import center_distance
from ..detection.constants import ATTRIBUTE_NAMES, DETECTION_NAMES, TP_METRICS


class DetectionConfig:
    """Data class that specifies the detection evaluation settings."""

    def __init__(
        self,
        class_range: Dict[str, int],
        dist_fcn: str,
        dist_ths: List[float],
        dist_th_tp: float,
        min_recall: float,
        min_precision: float,
        max_boxes_per_sample: int,
        mean_ap_weight: int,
    ):
        assert set(class_range.keys()) == set(DETECTION_NAMES), "Class count mismatch."
        assert dist_th_tp in dist_ths, "dist_th_tp must be in set of dist_ths."

        self.class_range = class_range
        self.dist_fcn = dist_fcn
        self.dist_ths = dist_ths
        self.dist_th_tp = dist_th_tp
        self.min_recall = min_recall
        self.min_precision = min_precision
        self.max_boxes_per_sample = max_boxes_per_sample
        self.mean_ap_weight = mean_ap_weight

        self.class_names = self.class_range.keys()

    def __eq__(self, other):
        eq = True
        for key in self.serialize().keys():
            eq = eq and np.array_equal(getattr(self, key), getattr(other, key))
        return eq

    def serialize(self) -> dict:
        """Serialize instance into json-friendly format."""
        return {
            "class_range": self.class_range,
            "dist_fcn": self.dist_fcn,
            "dist_ths": self.dist_ths,
            "dist_th_tp": self.dist_th_tp,
            "min_recall": self.min_recall,
            "min_precision": self.min_precision,
            "max_boxes_per_sample": self.max_boxes_per_sample,
            "mean_ap_weight": self.mean_ap_weight,
        }

    @classmethod
    def deserialize(cls, content: dict):
        """Initialize from serialized dictionary."""
        return cls(
            content["class_range"],
            content["dist_fcn"],
            content["dist_ths"],
            content["dist_th_tp"],
            content["min_recall"],
            content["min_precision"],
            content["max_boxes_per_sample"],
            content["mean_ap_weight"],
        )

    @property
    def dist_fcn_callable(self):
        """Return the distance function corresponding to the dist_fcn string."""
        if self.dist_fcn == "center_distance":
            return center_distance
        else:
            raise Exception("Error: Unknown distance function %s!" % self.dist_fcn)


class DetectionMetricData(MetricData):
    """This class holds accumulated and interpolated data required to calculate the detection metrics."""

    nelem = 101

    def __init__(
        self,
        recall: np.array,
        precision: np.array,
        confidence: np.array,
        trans_err: np.array,
        vel_err: np.array,
        scale_err: np.array,
        orient_err: np.array,
        attr_err: np.array,
    ):
        # Assert lengths.
        assert len(recall) == self.nelem
        assert len(precision) == self.nelem
        assert len(confidence) == self.nelem
        assert len(trans_err) == self.nelem
        assert len(vel_err) == self.nelem
        assert len(scale_err) == self.nelem
        assert len(orient_err) == self.nelem
        assert len(attr_err) == self.nelem

        # Assert ordering.
        assert all(
            confidence == sorted(confidence, reverse=True)
        )  # Confidences should be descending.
        assert all(recall == sorted(recall))  # Recalls should be ascending.

        # Set attributes explicitly to help IDEs figure out what is going on.
        self.recall = recall
        self.precision = precision
        self.confidence = confidence
        self.trans_err = trans_err
        self.vel_err = vel_err
        self.scale_err = scale_err
        self.orient_err = orient_err
        self.attr_err = attr_err

    def __eq__(self, other):
        eq = True
        for key in self.serialize().keys():
            eq = eq and np.array_equal(getattr(self, key), getattr(other, key))
        return eq

    @property
    def max_recall_ind(self):
        """Returns index of max recall achieved."""

        # Last instance of confidence > 0 is index of max achieved recall.
        non_zero = np.nonzero(self.confidence)[0]
        if len(non_zero) == 0:  # If there are no matches, all the confidence values will be zero.
            max_recall_ind = 0
        else:
            max_recall_ind = non_zero[-1]

        return max_recall_ind

    @property
    def max_recall(self):
        """Returns max recall achieved."""

        return self.recall[self.max_recall_ind]

    def serialize(self):
        """Serialize instance into json-friendly format."""
        return {
            "recall": self.recall.tolist(),
            "precision": self.precision.tolist(),
            "confidence": self.confidence.tolist(),
            "trans_err": self.trans_err.tolist(),
            "vel_err": self.vel_err.tolist(),
            "scale_err": self.scale_err.tolist(),
            "orient_err": self.orient_err.tolist(),
            "attr_err": self.attr_err.tolist(),
        }

    @classmethod
    def deserialize(cls, content: dict):
        """Initialize from serialized content."""
        return cls(
            recall=np.array(content["recall"]),
            precision=np.array(content["precision"]),
            confidence=np.array(content["confidence"]),
            trans_err=np.array(content["trans_err"]),
            vel_err=np.array(content["vel_err"]),
            scale_err=np.array(content["scale_err"]),
            orient_err=np.array(content["orient_err"]),
            attr_err=np.array(content["attr_err"]),
        )

    @classmethod
    def no_predictions(cls):
        """Returns a md instance corresponding to having no predictions."""
        return cls(
            recall=np.linspace(0, 1, cls.nelem),
            precision=np.zeros(cls.nelem),
            confidence=np.zeros(cls.nelem),
            trans_err=np.ones(cls.nelem),
            vel_err=np.ones(cls.nelem),
            scale_err=np.ones(cls.nelem),
            orient_err=np.ones(cls.nelem),
            attr_err=np.ones(cls.nelem),
        )

    @classmethod
    def random_md(cls):
        """Returns an md instance corresponding to a random results."""
        return cls(
            recall=np.linspace(0, 1, cls.nelem),
            precision=np.random.random(cls.nelem),
            confidence=np.linspace(0, 1, cls.nelem)[::-1],
            trans_err=np.random.random(cls.nelem),
            vel_err=np.random.random(cls.nelem),
            scale_err=np.random.random(cls.nelem),
            orient_err=np.random.random(cls.nelem),
            attr_err=np.random.random(cls.nelem),
        )


class DetectionMetrics:
    """Stores average precision and true positive metric results. Provides properties to summarize."""

    def __init__(self, cfg: DetectionConfig):
        self.cfg = cfg
        self._label_aps = defaultdict(lambda: defaultdict(float))
        self._label_tp_errors = defaultdict(lambda: defaultdict(float))
        self.eval_time = None

    def add_label_ap(self, detection_name: str, dist_th: float, ap: float) -> None:
        self._label_aps[detection_name][dist_th] = ap

    def get_label_ap(self, detection_name: str, dist_th: float) -> float:
        return self._label_aps[detection_name][dist_th]

    def add_label_tp(self, detection_name: str, metric_name: str, tp: float):
        self._label_tp_errors[detection_name][metric_name] = tp

    def get_label_tp(self, detection_name: str, metric_name: str) -> float:
        return self._label_tp_errors[detection_name][metric_name]

    def add_runtime(self, eval_time: float) -> None:
        self.eval_time = eval_time

    @property
    def mean_dist_aps(self) -> Dict[str, float]:
        """Calculates the mean over distance thresholds for each label."""
        return {class_name: np.mean(list(d.values())) for class_name, d in self._label_aps.items()}

    @property
    def mean_ap(self) -> float:
        """Calculates the mean AP by averaging over distance thresholds and classes."""
        return float(np.mean(list(self.mean_dist_aps.values())))

    @property
    def tp_errors(self) -> Dict[str, float]:
        """Calculates the mean true positive error across all classes for each metric."""
        errors = {}
        for metric_name in TP_METRICS:
            class_errors = []
            for detection_name in self.cfg.class_names:
                class_errors.append(self.get_label_tp(detection_name, metric_name))

            errors[metric_name] = float(np.nanmean(class_errors))

        return errors

    @property
    def tp_scores(self) -> Dict[str, float]:
        scores = {}
        tp_errors = self.tp_errors
        for metric_name in TP_METRICS:
            # We convert the true positive errors to "scores" by 1-error.
            score = 1.0 - tp_errors[metric_name]

            # Some of the true positive errors are unbounded, so we bound the scores to min 0.
            score = max(0.0, score)

            scores[metric_name] = score

        return scores

    @property
    def nd_score(self) -> float:
        """
        Compute the nuScenes detection score (NDS, weighted sum of the individual scores).
        :return: The NDS.
        """
        # Summarize.
        total = float(
            self.cfg.mean_ap_weight * self.mean_ap + np.sum(list(self.tp_scores.values()))
        )

        # Normalize.
        total = total / float(self.cfg.mean_ap_weight + len(self.tp_scores.keys()))

        return total

    def serialize(self):
        return {
            "label_aps": self._label_aps,
            "mean_dist_aps": self.mean_dist_aps,
            "mean_ap": self.mean_ap,
            "label_tp_errors": self._label_tp_errors,
            "tp_errors": self.tp_errors,
            "tp_scores": self.tp_scores,
            "nd_score": self.nd_score,
            "eval_time": self.eval_time,
            "cfg": self.cfg.serialize(),
        }

    @classmethod
    def deserialize(cls, content: dict):
        """Initialize from serialized dictionary."""

        cfg = DetectionConfig.deserialize(content["cfg"])

        metrics = cls(cfg=cfg)
        metrics.add_runtime(content["eval_time"])

        for detection_name, label_aps in content["label_aps"].items():
            for dist_th, ap in label_aps.items():
                metrics.add_label_ap(
                    detection_name=detection_name, dist_th=float(dist_th), ap=float(ap)
                )

        for detection_name, label_tps in content["label_tp_errors"].items():
            for metric_name, tp in label_tps.items():
                metrics.add_label_tp(
                    detection_name=detection_name, metric_name=metric_name, tp=float(tp)
                )

        return metrics

    def __eq__(self, other):
        eq = True
        eq = eq and self._label_aps == other._label_aps
        eq = eq and self._label_tp_errors == other._label_tp_errors
        eq = eq and self.eval_time == other.eval_time
        eq = eq and self.cfg == other.cfg

        return eq


class DetectionBox(EvalBox):
    """Data class used during detection evaluation. Can be a prediction or ground truth."""

    def __init__(
        self,
        sample_token: str = "",
        translation: Tuple[float, float, float] = (0, 0, 0),
        size: Tuple[float, float, float] = (0, 0, 0),
        rotation: Tuple[float, float, float, float] = (0, 0, 0, 0),
        velocity: Tuple[float, float] = (0, 0),
        ego_translation: Tuple[float, float, float] = (
            0,
            0,
            0,
        ),  # Translation to ego vehicle in meters.
        num_pts: int = -1,  # Nbr. LIDAR or RADAR inside the box. Only for gt boxes.
        detection_name: str = "car",  # The class name used in the detection challenge.
        detection_score: float = -1.0,  # GT samples do not have a score.
        attribute_name: str = "",
    ):  # Box attribute. Each box can have at most 1 attribute.
        super().__init__(
            sample_token,
            translation,
            size,
            rotation,
            velocity,
            ego_translation,
            num_pts,
        )

        assert detection_name is not None, "Error: detection_name cannot be empty!"
        assert detection_name in DETECTION_NAMES, (
            "Error: Unknown detection_name %s" % detection_name
        )

        assert attribute_name in ATTRIBUTE_NAMES or attribute_name == "", (
            "Error: Unknown attribute_name %s" % attribute_name
        )

        assert type(detection_score) == float, "Error: detection_score must be a float!"
        assert not np.any(np.isnan(detection_score)), "Error: detection_score may not be NaN!"

        # Assign.
        self.detection_name = detection_name
        self.detection_score = detection_score
        self.attribute_name = attribute_name

    def __eq__(self, other):
        return (
            self.sample_token == other.sample_token
            and self.translation == other.translation
            and self.size == other.size
            and self.rotation == other.rotation
            and self.velocity == other.velocity
            and self.ego_translation == other.ego_translation
            and self.num_pts == other.num_pts
            and self.detection_name == other.detection_name
            and self.detection_score == other.detection_score
            and self.attribute_name == other.attribute_name
        )

    def serialize(self) -> dict:
        """Serialize instance into json-friendly format."""
        return {
            "sample_token": self.sample_token,
            "translation": self.translation,
            "size": self.size,
            "rotation": self.rotation,
            "velocity": self.velocity,
            "ego_translation": self.ego_translation,
            "num_pts": self.num_pts,
            "detection_name": self.detection_name,
            "detection_score": self.detection_score,
            "attribute_name": self.attribute_name,
        }

    @classmethod
    def deserialize(cls, content: dict):
        """Initialize from serialized content."""
        return cls(
            sample_token=content["sample_token"],
            translation=tuple(content["translation"]),
            size=tuple(content["size"]),
            rotation=tuple(content["rotation"]),
            velocity=tuple(content["velocity"]),
            ego_translation=(0.0, 0.0, 0.0)
            if "ego_translation" not in content
            else tuple(content["ego_translation"]),
            num_pts=-1 if "num_pts" not in content else int(content["num_pts"]),
            detection_name=content["detection_name"],
            detection_score=-1.0
            if "detection_score" not in content
            else float(content["detection_score"]),
            attribute_name=content["attribute_name"],
        )


class DetectionMetricDataList:
    """This stores a set of MetricData in a dict indexed by (name, match-distance)."""

    def __init__(self):
        self.md = {}

    def __getitem__(self, key):
        return self.md[key]

    def __eq__(self, other):
        eq = True
        for key in self.md.keys():
            eq = eq and self[key] == other[key]
        return eq

    def get_class_data(self, detection_name: str) -> List[Tuple[DetectionMetricData, float]]:
        """Get all the MetricData entries for a certain detection_name."""
        return [(md, dist_th) for (name, dist_th), md in self.md.items() if name == detection_name]

    def get_dist_data(self, dist_th: float) -> List[Tuple[DetectionMetricData, str]]:
        """Get all the MetricData entries for a certain match_distance."""
        return [
            (md, detection_name)
            for (detection_name, dist), md in self.md.items()
            if dist == dist_th
        ]

    def set(self, detection_name: str, match_distance: float, data: DetectionMetricData):
        """Sets the MetricData entry for a certain detection_name and match_distance."""
        self.md[(detection_name, match_distance)] = data

    def serialize(self) -> dict:
        return {key[0] + ":" + str(key[1]): value.serialize() for key, value in self.md.items()}

    @classmethod
    def deserialize(cls, content: dict):
        mdl = cls()
        for key, md in content.items():
            name, distance = key.split(":")
            mdl.set(name, float(distance), DetectionMetricData.deserialize(md))
        return mdl



"""Calibration dataclasses."""
import json
from dataclasses import dataclass
from typing import Any, Dict

import numpy as np

from zod.constants import EGO, Camera, CoordinateFrame, Lidar
from zod.data_classes.geometry import Pose
from zod.utils.geometry import transform_points


@dataclass
class LidarCalibration:
    extrinsics: Pose  # lidar pose in the ego frame


@dataclass
class CameraCalibration:
    extrinsics: Pose  # 4x4 matrix describing the camera pose in the ego frame
    intrinsics: np.ndarray  # 3x3 matrix
    distortion: np.ndarray  # 4 vector
    undistortion: np.ndarray  # 4 vector
    image_dimensions: np.ndarray  # width, height
    field_of_view: np.ndarray  # horizontal, vertical (degrees)


@dataclass
class Calibration:
    lidars: Dict[Lidar, LidarCalibration]
    cameras: Dict[Camera, CameraCalibration]

    @classmethod
    def from_dict(cls, calib_dict: Dict[str, Any]) -> "Calibration":
        lidars = {
            Lidar.VELODYNE: LidarCalibration(
                extrinsics=Pose(np.array(calib_dict["FC"]["lidar_extrinsics"]))
            ),
        }
        cameras = {
            Camera.FRONT: CameraCalibration(
                extrinsics=Pose(np.array(calib_dict["FC"]["extrinsics"])),
                intrinsics=np.array(calib_dict["FC"]["intrinsics"]),
                distortion=np.array(calib_dict["FC"]["distortion"]),
                undistortion=np.array(calib_dict["FC"]["undistortion"]),
                image_dimensions=np.array(calib_dict["FC"]["image_dimensions"]),
                field_of_view=np.array(calib_dict["FC"]["field_of_view"]),
            ),
        }
        return cls(lidars=lidars, cameras=cameras)

    @classmethod
    def from_json_path(cls, json_path: str) -> "Calibration":
        with open(json_path) as f:
            calib_dict = json.load(f)
        return cls.from_dict(calib_dict)

    def transform_points(
        self,
        points: np.ndarray,
        from_frame: CoordinateFrame,
        to_frame: CoordinateFrame,
    ) -> np.ndarray:
        """Transform points from one frame to another.

        Args:
            points: Points in shape (N, 3).
            from_frame: Name of the frame the points are in.
            to_frame: Name of the frame to transform the points to.

        Returns:
            Transformed points in shape (N, 3).
        """
        if from_frame == to_frame:
            return points
        from_transform = self.get_extrinsics(from_frame).transform
        to_transform = self.get_extrinsics(to_frame).transform
        return transform_points(points, to_transform @ from_transform)

    def get_extrinsics(self, frame: CoordinateFrame) -> Pose:
        if frame == EGO:
            return Pose.identity()
        elif isinstance(frame, Lidar):
            return self.lidars[frame].extrinsics
        elif isinstance(frame, Camera):
            return self.cameras[frame].extrinsics
        else:
            raise ValueError(f"Unknown frame {frame}")



from dataclasses import dataclass
from datetime import datetime

from ._serializable import JSONSerializable


@dataclass
class FrameMetaData(JSONSerializable):
    """A class describing the metadata of a frame."""

    frame_id: str
    time: datetime
    country_code: str
    scraped_weather: str
    collection_car: str
    road_type: str
    road_condition: str
    time_of_day: str
    num_lane_instances: int
    num_vehicles: int
    num_vulnerable_vehicles: int
    num_pedestrians: int
    num_traffic_lights: int
    num_traffic_signs: int
    longitude: float
    latitude: float
    solar_angle_elevation: float


@dataclass
class SequenceMetadata(JSONSerializable):
    """A class describing the metadata of a sequence."""

    sequence_id: str
    start_time: datetime
    end_time: datetime
    country_code: str
    collection_car: str
    longitude: float
    latitude: float



from typing import Any, List, Optional

import numpy as np

from zod.constants import AnnotationProject, Anonymization, Camera, Lidar
from zod.utils.compensation import motion_compensate_scanwise

from .calibration import Calibration
from .info import Information
from .metadata import FrameMetaData
from .oxts import EgoMotion
from .sensor import CameraFrame, LidarData, LidarFrame


class ZodFrame:
    def __init__(self, info: Information):
        self.info: Information = info  # Holds all the paths to the files
        self._ego_motion: EgoMotion = None  # This is the lightweight version of oxts
        self._oxts: EgoMotion = None
        self._calibration: Calibration = None
        self._metadata: FrameMetaData = None

    @property
    def ego_motion(self) -> EgoMotion:
        """Get the ego motion."""
        if self._ego_motion is None:
            self._ego_motion = EgoMotion.from_json_path(self.info.ego_motion_path)
        return self._ego_motion

    @property
    def oxts(self) -> EgoMotion:
        """Get the oxts."""
        if self._oxts is None:
            self._oxts = EgoMotion.from_oxts_path(self.info.oxts_path)
        return self._oxts

    @property
    def calibration(self) -> Calibration:
        """Get the calibration."""
        if self._calibration is None:
            self._calibration = Calibration.from_json_path(self.info.calibration_path)
        return self._calibration

    @property
    def metadata(self) -> FrameMetaData:
        """Get the metadata."""
        if self._metadata is None:
            self._metadata = FrameMetaData.from_json_path(self.info.metadata_path)
        return self._metadata

    def get_annotation(self, project: AnnotationProject) -> List[Any]:
        """Get the annotation for a given project."""
        assert project in self.info.annotation_frames, f"Project {project} not available."
        return self.info.get_key_annotation_frame(project).read()

    def get_camera_frame(self, anonymization: Anonymization = Anonymization.BLUR) -> CameraFrame:
        """Get the camera frame."""
        return self.info.get_key_camera_frame(camera=Camera.FRONT, anonymization=anonymization)

    def get_image(
        self,
        anonymization: Anonymization = Anonymization.BLUR,
    ) -> np.ndarray:
        """Get the image."""
        return self.info.get_key_camera_frame(
            camera=Camera.FRONT, anonymization=anonymization
        ).read()

    def get_lidar_frames(
        self,
        num_before: int = 0,
        num_after: int = 0,
    ) -> List[LidarFrame]:
        """Get the lidar frames (around the keyframe)."""
        all_frames = self.info.get_lidar_frames(Lidar.VELODYNE)
        key_frame_dx = len(all_frames) // 2  # the key frame is in the middle
        start = max(0, key_frame_dx - num_before)
        end = min(len(all_frames), key_frame_dx + num_after + 1)
        return all_frames[start:end]

    def get_lidar(self, num_before: int = 0, num_after: int = 0) -> List[LidarData]:
        """Get the lidar data, same as `get_lidar_frames` but actually reads the data."""
        return [lidar_frame.read() for lidar_frame in self.get_lidar_frames(num_before, num_after)]

    def compensate_lidar(self, data: LidarData, timestamp: float) -> LidarData:
        """Compensate a point cloud to a given timestamp."""
        lidar_calib = self.calibration.lidars[Lidar.VELODYNE]
        return motion_compensate_scanwise(data, self.ego_motion, lidar_calib, timestamp)

    def get_aggregated_lidar(
        self, num_before: int, num_after: int = 0, timestamp: Optional[float] = None
    ) -> LidarData:
        """Get an aggregated point cloud around the keyframe."""
        key_lidar_frame = self.info.get_key_lidar_frame()
        key_lidar_data = key_lidar_frame.read()
        _adjust_lidar_core_time(key_lidar_data)
        if timestamp is None:
            timestamp = key_lidar_data.core_timestamp
        lidar_calib = self.calibration.lidars[Lidar.VELODYNE]
        # Adjust each individual scan
        to_aggregate = []
        for lidar_frame in self.get_lidar_frames(num_before, num_after):
            if lidar_frame == key_lidar_frame:
                continue
            lidar_data = lidar_frame.read()
            _adjust_lidar_core_time(lidar_data)
            lidar_data = motion_compensate_scanwise(
                lidar_data, self.ego_motion, lidar_calib, timestamp
            )
            to_aggregate.append(lidar_data)
        # Aggregate the scans
        key_lidar_data.extend(*to_aggregate)
        return key_lidar_data


def _adjust_lidar_core_time(lidar: LidarData):
    """Adjust the core timestamp of a lidar scan so that it always points "forward".

    This assumes that the scans are not pointwise compensated, and that the cut-off is to the right.
    """
    # TODO: maybe this should be done by looking at the angles instead?
    # This could fail if there are no points on either side of the cut-off.
    lidar.core_timestamp = 0.75 * lidar.timestamps.max() + 0.25 * lidar.timestamps.min()



"""This module contains functions for matching detections and ground truth.

As input we have a list of ground truth objects and a list of predictions.
The ground truth objects contains objects that correspond to dont-care regions,
for example parking lots, bike racks, etc. They can also be objects that does
not have any 3D properties (i.e., no 3D bounding box). This can be the case
for very distant objects that are visible in the camera but not in the LiDAR.
All these type of ground truth objects are ignored in the metric computation.

They are ignored by matching 3D false positives to these objects based on the
projection of the 3D bounding box into the image plane. These are matched with
ignore/dont-care objects by a IoU threshold.

The general outline of the matching process is as follows:
- Match the detections to the ground truth in 3D based on some metric.
- For all false positives in the 3D matching, we check the Birdeye View (BEV)
    intersection over detection-area (IoD) metric with all dont-care frustums
    (that also have been projected from camera 2D -> 3D -> BEV). Those false
    positives that have an IoD > over a certain threshold with a dont-care
    are removed from the false positives.
- Compute the relevant metrics based on the 3D matching and the
    false positives/false negatives.
"""

from dataclasses import dataclass
from typing import Callable, List, Tuple

import numpy as np
from scipy.optimize import linear_sum_assignment
from shapely import geometry

from zod.anno.object import AnnotatedObject, PredictedObject
from zod.constants import EVALUATION_FRAME
from zod.data_classes import Calibration
from zod.eval.detection.utils import polygon_iod2D

GtPredMatch = Tuple[AnnotatedObject, PredictedObject]
MIN_CAMERA_ONLY_DEPTH = 150  # in meters
MAX_CAMERA_ONLY_DEPTH = 500  # in meters


@dataclass
class MatchedFrame:
    """Container for matching results."""

    matches: List[GtPredMatch]
    false_positives: List[PredictedObject]
    false_negatives: List[AnnotatedObject]


def split_gt_objects(
    ground_truth: List[AnnotatedObject],
) -> Tuple[List[AnnotatedObject], List[AnnotatedObject]]:
    """Split the ground truth objects into valid and dont-care objects."""

    # valid ground truth objects
    valid_gt: List[AnnotatedObject] = []
    # dont-care ground truth objects
    dont_care_gt: List[AnnotatedObject] = []
    for gt in ground_truth:
        if gt.should_ignore_object():
            dont_care_gt.append(gt)
        else:
            valid_gt.append(gt)

    return valid_gt, dont_care_gt


def match_dont_care_objects(
    dont_care_gt: List[AnnotatedObject],
    false_positives: List[PredictedObject],
    calibration: Calibration,
    fp_dist_fcn: Callable = polygon_iod2D,
    fp_dist_threshold: float = 0.25,
) -> List[PredictedObject]:
    """Match the false positives to the dont-care objects."""

    dont_care_polygons: List[geometry.Polygon] = []
    # turn all into frustums
    for gt in dont_care_gt:
        # if we have the 3d properties we can use the 3D box instead
        # of the frustum created from the 2D box.
        if gt.box3d is not None:
            bev_corners = gt.box3d.corners_bev
            bev_polygon = geometry.Polygon(bev_corners)
            dont_care_polygons.append(bev_polygon)
            continue
        # if we dont have 3d properties we use the 2D box
        # to create a frustum
        frustum = gt.box2d.get_3d_frustum(
            calibration,
            frame=EVALUATION_FRAME,
            min_depth=MIN_CAMERA_ONLY_DEPTH,
            max_depth=MAX_CAMERA_ONLY_DEPTH,
        )
        # ignore the z dimension as we move to bev
        frustum = frustum[:, :2]
        inner = frustum[:4]
        outer = frustum[4:]

        # initialize the polygon
        polygon_points = np.zeros((4, 2))
        # get the index of the minimum x value
        min_x_idx = np.argmin(inner[:, 0])
        # get the index of the maximum x value
        max_x_idx = np.argmax(inner[:, 0])
        # get the index of the minimum x value for the outer
        min_x_idx_outer = np.argmin(outer[:, 0])
        # get the index of the maximum x value for the outer
        max_x_idx_outer = np.argmax(outer[:, 0])
        polygon_points[0] = inner[min_x_idx]
        polygon_points[1] = inner[max_x_idx]
        polygon_points[2] = outer[max_x_idx_outer]
        polygon_points[3] = outer[min_x_idx_outer]

        dont_care_polygons.append(geometry.Polygon(polygon_points))

    match_to_dont_care: List[PredictedObject] = []

    for pred in false_positives:
        bev_corners = pred.box3d.corners_bev
        pred_poly = geometry.Polygon(bev_corners)

        for dc_polygon in dont_care_polygons:
            dist = fp_dist_fcn(pred_poly, dc_polygon)
            if dist > fp_dist_threshold:
                match_to_dont_care.append(pred)
                break

    return match_to_dont_care


def greedy_match(
    ground_truth: List[AnnotatedObject],
    predictions: List[PredictedObject],
    calibration: Calibration,
    dist_fcn: Callable,
    dist_threshold: float,
) -> MatchedFrame:
    """This function will greedily match the detections to the ground truth.

    It will use the confifence score to determine the order of the detections.

    Returns:
        - A list of matched ground truth objects and predictions.
        - A list of unmatched predictions.
        - A list of unmatched ground truth objects.
    """
    # trivial cases
    if len(ground_truth) == 0:
        return MatchedFrame([], predictions, [])
    if len(predictions) == 0:
        return MatchedFrame([], [], ground_truth)
    if len(ground_truth) == len(predictions) == 0:
        return MatchedFrame([], [], [])

    # Sort the predictions by their confidence score.
    predictions = sorted(predictions, key=lambda x: x.confidence, reverse=True)

    valid_gt, dont_care_gt = split_gt_objects(ground_truth)

    # Create a list of tuples. Each tuple contains the ground truth object and the prediction.
    matches: List[GtPredMatch] = []
    unmatched_predictions: List[PredictedObject] = []
    # Iterate over the predictions.
    for pred in predictions:
        # Iterate over all the ground truth objects and find the best match.
        min_dist = np.inf
        best_match_idx = None
        for gt_idx, gt in enumerate(valid_gt):
            distance = dist_fcn(gt.box3d, pred.box3d)
            if distance < min_dist:
                min_dist = distance
                best_match_idx = gt_idx

        if min_dist > dist_threshold:
            unmatched_predictions.append(pred)
        else:
            # We have a match.
            matches.append((valid_gt.pop(best_match_idx), pred))

    if len(unmatched_predictions) > 0:
        dont_care_matches = match_dont_care_objects(
            dont_care_gt, unmatched_predictions, calibration
        )

        # Remove the matched false positives from the unmatched predictions.
        unmatched_predictions = [
            pred for pred in unmatched_predictions if (pred not in dont_care_matches)
        ]

    # return the matches, the unmatched predictions and the remaining valid ground truth objects
    return MatchedFrame(
        matches=matches,
        false_positives=unmatched_predictions,
        false_negatives=valid_gt,
    )


def optimal_match(
    ground_truth: List[AnnotatedObject],
    predictions: List[PredictedObject],
    calibration: Calibration,
    dist_fcn: Callable,
    dist_threshold: float,
) -> MatchedFrame:
    """This function will make an optimal matching between the detections and the ground truth.

    Use the Hungarian algorithm to find the optimal matching.
    """

    if len(ground_truth) == 0:
        return MatchedFrame([], predictions, [])
    if len(predictions) == 0:
        return MatchedFrame([], [], ground_truth)
    if len(ground_truth) == len(predictions) == 0:
        return MatchedFrame([], [], [])

    if len(ground_truth):
        # assert that all ground truth objects are in the same frame
        assert len(set([gt.box3d.frame for gt in ground_truth])) == 1
    if len(predictions):
        # assert that all predictions are in the same frame
        assert len(set([pred.box3d.frame for pred in predictions])) == 1
    if len(ground_truth) and len(predictions):
        # assert that the frames are the same
        assert ground_truth[0].box3d.frame == predictions[0].box3d.frame

    valid_gt, dont_care_gt = split_gt_objects(ground_truth)

    # initialize the cost matrix
    cost_matrix = np.zeros((len(valid_gt), len(predictions)))
    for gt_idx, gt in enumerate(ground_truth):
        for pred_idx, pred in enumerate(predictions):
            dist = dist_fcn(gt.box3d, pred.box3d)
            if dist > dist_threshold:
                cost_matrix[gt_idx, pred_idx] = np.inf
            else:
                cost_matrix[gt_idx, pred_idx] = -(dist + np.log(pred.confidence))

    # find the optimal matching
    row_ind, col_ind = linear_sum_assignment(cost_matrix)

    # create the matches
    matches: List[GtPredMatch] = []
    unmatched_predictions: List[PredictedObject] = []
    for gt_idx, pred_idx in zip(row_ind, col_ind):
        if cost_matrix[gt_idx, pred_idx] == np.inf:
            unmatched_predictions.append(predictions[pred_idx])
        else:
            matches.append((valid_gt[gt_idx], predictions[pred_idx]))

    match_to_dont_care = match_dont_care_objects(dont_care_gt, unmatched_predictions, calibration)

    # Remove the matched false positives from the unmatched predictions.
    unmatched_predictions = [
        pred for pred in unmatched_predictions if pred not in match_to_dont_care
    ]

    # return the matches, the unmatched predictions and the remaining valid ground truth objects
    return MatchedFrame(
        matches=matches,
        false_positives=unmatched_predictions,
        false_negatives=valid_gt,
    )


def match_one_frame(
    ground_truth: List[AnnotatedObject],
    predictions: List[PredictedObject],
    calibration: Calibration,
    dist_fcn: Callable,
    dist_threshold: float,
    method: str = "greedy",
) -> MatchedFrame:
    """Match the detections to the ground truth.

    Args:
        ground_truth: A list of ground truth objects.
        predictions: A list of predictions.
        calibration: The calibration of the scene.
        dist_fcn: The distance function to use. should take 2 Box3D objects as arguments
        dist_threshold: The distance threshold.
        method: The method to use. Can be "greedy" or "optimal".
    Returns:
        A list of tuples. Each tuple contains the ground truth object and the prediction.
    """
    # check that all objects are in the evaluation frame
    assert all(
        gt.box3d.frame == EVALUATION_FRAME for gt in ground_truth if gt.box3d is not None
    ), "All ground truth objects must be in the evaluation frame."
    assert all(
        pred.box3d.frame == EVALUATION_FRAME for pred in predictions
    ), "All predictions must be in the evaluation frame."

    if method == "greedy":
        return greedy_match(
            ground_truth=ground_truth,
            predictions=predictions,
            calibration=calibration,
            dist_fcn=dist_fcn,
            dist_threshold=dist_threshold,
        )
    elif method == "optimal":
        return optimal_match(
            ground_truth=ground_truth,
            predictions=predictions,
            calibration=calibration,
            dist_fcn=dist_fcn,
            dist_threshold=dist_threshold,
        )
    else:
        raise ValueError(f"Unknown method: {method}")



from .polygon_utils import overlay_polygon_on_image


def overlay_ego_road_on_image(polygons, image, fill_color=(0, 100, 0), alpha=1.0):
    for polygon in polygons:
        image = overlay_polygon_on_image(polygon, image, fill_color, alpha)
    return image



"""Util to plot BEV."""
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
import plotly.express as px
import plotly.graph_objects as go

from .bev_utils import BEVSettings, create_pointcloud_input


class BEVBox:
    """Write objects on top of a discrete bird's eye view tensor."""

    def __init__(self):
        """Construct instance."""
        super().__init__()
        self._settings = BEVSettings()
        # Exclude dark gray color
        self._color = px.colors.qualitative.Dark24[:5] + px.colors.qualitative.Dark24[6:]

    def __call__(
        self, bev: np.ndarray, objects: Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
    ) -> go.Figure:
        """Go through all objects for each batch, and write them on top the BEV."""
        input_ = create_pointcloud_input(bev, self._settings)
        classes, positions, dimensions, rotations = objects
        n_objects = np.shape(positions)[0]

        # Setup figure
        fig = self._setup_figure()

        # Plot the reduced BEV
        input_ = self._create_od_vis_background(input_)
        fig.add_trace(go.Image(z=input_))

        # Add object rectangles, if any
        if n_objects > 0:
            # Map classes
            classes = self._settings.encode_classes(classes)
            # Scale dimensions and positions
            dimensions = dimensions[:, :2] / self._settings.grid_cell_size
            positions = (positions[:, :2] - self._settings.grid_min) / self._settings.grid_cell_size
            for obj_idx in range(n_objects):
                self._add_object(
                    fig,
                    classes[obj_idx],
                    positions[obj_idx],
                    dimensions[obj_idx],
                    rotations[obj_idx],
                )
        self._activate_legend(fig)
        fig.show()

        return fig

    def _setup_figure(self) -> go.Figure:
        limits = (self._settings.grid_max - self._settings.grid_min) / self._settings.grid_cell_size
        # Setup a figure with correct scaling and ticks.
        xtickvals, xticktext, ytickvals, yticktext = self._calculate_ticks(limits)
        layout = go.Layout(
            width=1000,
            height=1000,
            xaxis=dict(
                range=[0, int(limits[0])],
                title="Lateral (meters)",
                tickvals=xtickvals,
                ticktext=xticktext,
            ),
            yaxis=dict(
                range=[0, int(limits[1])],
                title="Longitudinal (meters)",
                tickvals=ytickvals,
                ticktext=yticktext,
                scaleanchor="x",
            ),
        )
        fig = go.Figure(layout=layout)
        return fig

    def _calculate_ticks(
        self, limits: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        _, ax = plt.subplots(1, 1)
        ax.set_aspect("equal")
        ax.set_xlim(0, int(limits[0]))
        ax.set_ylim(0, int(limits[1]))
        x_ticks = ax.get_xticks()
        y_ticks = ax.get_yticks()
        x_ticks_text = [
            tick * self._settings.grid_cell_size + self._settings.grid_min[0] for tick in x_ticks
        ]
        y_ticks_text = [
            tick * self._settings.grid_cell_size + self._settings.grid_min[1] for tick in y_ticks
        ]
        plt.close()
        return x_ticks, x_ticks_text, y_ticks, y_ticks_text

    def _add_object(
        self,
        fig: go.Figure,
        class_ind: int,
        position: np.ndarray,
        dimension: np.ndarray,
        rotation: np.ndarray,
    ):
        # pylint: disable=too-many-arguments
        # Extract object information
        coord1 = (-1) * dimension / 2
        coord3 = dimension / 2
        coord2 = [coord1[0], coord3[1]]
        coord4 = [coord3[0], coord1[1]]

        all_coords = np.vstack([coord1, coord2, coord3, coord4])
        all_coords = np.hstack([all_coords, np.zeros((4, 1))])

        all_coords = [rotation.rotate(point)[:2] + position for point in all_coords]
        all_coords = np.vstack([all_coords, all_coords[0]])

        # Add rotated bounding box
        fig.add_scatter(
            x=all_coords[:, 0],
            y=all_coords[:, 1],
            line=dict(color=self._color[class_ind], width=self._settings.grid_cell_size * 30),
            marker=dict(size=self._settings.grid_cell_size * 30),
            legendgroup=self._settings.get_class_name(class_ind),
            name=self._settings.get_class_name(class_ind),
            showlegend=False,
        )

        # Add arrow in the direction of the object
        arrow_length = int(np.cast["float32"](dimension[0]) * 0.8)
        end_point = np.array([arrow_length, 0, 0])
        end_point = rotation.rotate(end_point)[:2] + position
        fig.add_annotation(
            x=end_point[0],
            y=end_point[1],
            xref="x",
            yref="y",
            text="",
            showarrow=True,
            axref="x",
            ayref="y",
            ax=position[0],
            ay=position[1],
            arrowhead=3,
            arrowwidth=self._settings.grid_cell_size * 20,
            arrowcolor="#247BA0",
        )

    def _activate_legend(self, fig):
        """Activate legend with all classes."""
        for idx, _ in enumerate(self._settings.classes):
            fig.add_scatter(
                x=[0],
                y=[0],
                legendgroup=self._settings.get_class_name(idx),
                name=self._settings.get_class_name(idx),
                showlegend=True,
                marker=dict(color=self._color[idx], size=0.01),
            )

    @staticmethod
    def _create_od_vis_background(input_array: np.ndarray) -> np.ndarray:
        """Create a gray occupancy grid as background to visualize over."""
        occupancy = np.maximum.reduce(
            np.cast["float32"](np.abs(input_array) > 0.0), axis=0, keepdims=True
        )
        vis_bg = np.transpose(np.repeat(occupancy * 77, 3, axis=0), [2, 1, 0])
        return vis_bg



import os.path as osp
from dataclasses import dataclass
from datetime import datetime
from itertools import chain
from typing import Dict, Iterator, List, Tuple

from zod.constants import AnnotationProject, Anonymization, Camera, Lidar

from ._serializable import JSONSerializable
from .sensor import AnnotationFrame, CameraFrame, LidarFrame, SensorFrame


@dataclass
class Information(JSONSerializable):
    """Base class for frame and sequence information."""

    id: str
    start_time: datetime
    end_time: datetime
    keyframe_time: datetime

    calibration_path: str
    ego_motion_path: str
    metadata_path: str
    oxts_path: str

    annotation_frames: Dict[AnnotationProject, List[AnnotationFrame]]
    camera_frames: Dict[str, List[CameraFrame]]  # key is a combination of Camera and Anonymization
    lidar_frames: Dict[Lidar, List[LidarFrame]]

    @property
    def all_frames(self) -> Iterator[SensorFrame]:
        """Iterate over all frames."""
        return chain(
            *self.lidar_frames.values(),
            *self.camera_frames.values(),
            *self.annotation_frames.values(),
        )

    def convert_paths_to_absolute(self, root_path: str):
        self.calibration_path = osp.join(root_path, self.calibration_path)
        self.ego_motion_path = osp.join(root_path, self.ego_motion_path)
        self.metadata_path = osp.join(root_path, self.metadata_path)
        self.oxts_path = osp.join(root_path, self.oxts_path)
        for frame in self.all_frames:
            frame.filepath = osp.join(root_path, frame.filepath)

    def get_camera_lidar_map(
        self, anonymization: Anonymization, camera: Camera, lidar: Lidar
    ) -> Iterator[Tuple[CameraFrame, SensorFrame]]:
        """Iterate over all camera frames and their corresponding lidar frames.

        Args:
            camera: The camera to use. e.g., camera_front_blur
            lidar: The lidar to use. e.g., lidar_velodyne
        Yields:
            A tuple of the camera frame and the closest lidar frame.
        """
        camera_name = f"{camera.value}_{anonymization.value}"
        assert (
            camera_name in self.camera_frames
        ), f"Camera {camera_name} not found. Available cameras: {self.camera_frames.keys()}"
        assert (
            lidar in self.lidar_frames
        ), f"Lidar {lidar} not found. Available lidars: {self.lidar_frames.keys()}"

        for camera_frame in self.camera_frames[camera_name]:
            # Get the closest lidar frame in time
            lidar_frame = min(
                self.lidar_frames[lidar],
                key=lambda lidar_frame: abs(lidar_frame.time - camera_frame.time),
            )
            yield camera_frame, lidar_frame

    ### Keyframe accessors ###

    def get_key_annotation_frame(self, project: AnnotationProject) -> AnnotationFrame:
        if project not in self.annotation_frames:
            raise ValueError(f"No annotations found for project {project.value}.")
        if len(self.annotation_frames[project]) == 1:
            return self.annotation_frames[project][0]
        else:
            return self.get_annotation_frame(self.keyframe_time, project)

    def get_key_camera_frame(
        self,
        anonymization: Anonymization,
        camera: Camera = Camera.FRONT,
    ) -> CameraFrame:
        camera_name = f"{camera.value}_{anonymization.value}"
        if len(self.camera_frames[camera_name]) == 1:
            return self.camera_frames[camera_name][0]
        else:
            return min(
                self.camera_frames[camera_name],
                key=lambda camera_frame: abs(camera_frame.time - self.keyframe_time),
            )

    def get_key_lidar_frame(self, lidar: Lidar = Lidar.VELODYNE) -> LidarFrame:
        return self.get_lidar_frame(self.keyframe_time, lidar)

    ### Timestamp accessors ###

    def get_annotation_frame(
        self,
        time: datetime,
        project: AnnotationProject,
        # TODO: implement exact flag?
    ) -> AnnotationFrame:
        if project not in self.annotation_frames:
            raise ValueError(f"No annotations found for project {project.value}.")
        return min(
            self.annotation_frames[project],
            key=lambda annotation_frame: abs(annotation_frame.time - time),
        )

    def get_camera_frame(
        self,
        time: datetime,
        anonymization: Anonymization,
        camera: Camera = Camera.FRONT,
    ) -> CameraFrame:
        camera_name = f"{camera.value}_{anonymization.value}"
        return min(
            self.camera_frames[camera_name],
            key=lambda camera_frame: abs(camera_frame.time - time),
        )

    def get_lidar_frame(self, time: datetime, lidar: Lidar = Lidar.VELODYNE) -> LidarFrame:
        return min(
            self.lidar_frames[lidar],
            key=lambda lidar_frame: abs(lidar_frame.time - time),
        )

    ### Full accessors ###

    def get_annotation_frames(self, project: AnnotationProject) -> List[AnnotationFrame]:
        if project not in self.annotation_frames:
            return []
        return self.annotation_frames[project]

    def get_camera_frames(
        self,
        anonymization: Anonymization,
        camera: Camera = Camera.FRONT,
    ) -> List[CameraFrame]:
        camera_name = f"{camera.value}_{anonymization.value}"
        return self.camera_frames[camera_name]

    def get_lidar_frames(self, lidar: Lidar = Lidar.VELODYNE) -> List[LidarFrame]:
        return self.lidar_frames[lidar]



from pathlib import Path
from typing import Tuple

try:
    import typer
except ImportError:
    print('Warning! zod is installed without the CLI dependencies:\npip install "zod[cli]"')
    exit(1)

from zod.cli.download_zod import app as download_app
from zod.cli.generate_coco_json import convert_to_coco
from zod.cli.visualize_lidar import app as visualize_lidar_app

app = typer.Typer(help="Zenseact Open Dataset CLI.", no_args_is_help=True)

app.add_typer(download_app, name="download")

visualize_app = typer.Typer(help="Visualization tools", no_args_is_help=True)
app.add_typer(visualize_app, name="visualize")
visualize_app.add_typer(visualize_lidar_app, name="lidar")


convert_app = typer.Typer(
    help="Convert the Zenseact Open Dataset to a different format.", no_args_is_help=True
)


def tsr_dummy(
    dataset_root: Path = typer.Option(
        ...,
        exists=True,
        dir_okay=True,
        writable=False,
        readable=True,
        resolve_path=True,
        help="Path to the root of the ZOD dataset.",
    ),
    output_dir: Path = typer.Option(
        ...,
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=True,
        help="Path to the output directory.",
    ),
    version: str = typer.Option("full", help="Version of the dataset to use. One of: full, small."),
    path_size: Tuple[int, int] = typer.Option((64, 64), help="Path resultion."),
):
    typer.echo("Not Implemented Yet")


convert_app.command("coco", no_args_is_help=True)(convert_to_coco)
convert_app.command("tsr-patches", no_args_is_help=True)(tsr_dummy)
app.add_typer(convert_app, name="generate")


if __name__ == "__main__":
    app()
